# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional
import pandas as pd
import torch
from torch.utils.data import Dataset


@dataclass
class TimeSeries:
    """
    A time series contains the time series data along with metadata about the
    time series as well as static and dynamic features.
    """

    dataset_name: str
    start_date: Optional[pd.Timestamp]
    values: torch.Tensor
    item_id: Optional[str] = None
    feat_static_cat: Optional[torch.Tensor] = None
    scale: Optional[torch.Tensor] = None

    # ---------------------------------------------------------------------------------------------
    @property
    def end_date(self) -> pd.Timestamp:
        return self.start_date + (self.__len__() - 1) * self.start_date.freq

    @property
    def mean(self) -> torch.Tensor:
        return torch.mean(self.values, dim=0)

    @property
    def std(self) -> torch.Tensor:
        return torch.std(self.values, dim=0)

    def standardize(self, m: torch.Tensor, std: torch.Tensor):
        return TimeSeries(
            dataset_name=self.dataset_name,
            item_id=self.item_id,
            start_date=self.start_date,
            values=(self.values - m) / std,
            feat_static_cat=self.feat_static_cat,
            scale=torch.cat([m, std]),
        )

    def __len__(self) -> int:
        return self.values.shape[0]

    def __getitem__(self, sequence: slice) -> TimeSeries:
        assert (
            sequence.start >= 0
        ), "Time series cannot be sliced prior to start."
        assert (
            sequence.step is None or sequence.step == 1
        ), "Time series cannot be sliced with a step size other than 1."

        return TimeSeries(
            dataset_name=self.dataset_name,
            item_id=self.item_id,
            start_date=self.start_date + sequence.start * self.start_date.freq,
            values=self.values[sequence],
            feat_static_cat=self.feat_static_cat,
            scale=self.scale,
        )


class TimeSeriesDataset(Dataset[TimeSeries]):
    """
    A dataset which provides time series.
    """

    def __init__(
        self,
        series: List[TimeSeries],
        prediction_length: int,
        freq: str,
        standardize: bool = True,
    ):
        """
        Args:
            series: The multivariate time series.
        """
        self.series = series
        self.prediction_length = prediction_length
        self.freq = freq
        self.standardize = standardize
        if self.standardize:
            self.means = torch.stack([s.mean for s in self.series], dim=0)
            self.stds = torch.stack([s.std for s in self.series], dim=0)

    def rescale_dataset(self, series: torch.Tensor):
        """
        Redo standardization.

        The series must contain the same time series in the same order as the
        dataset.
        """
        return (
            (series * self.stds.unsqueeze(2)) + self.means.unsqueeze(2)
            if self.standardize
            else series
        )

    @property
    def number_of_time_steps(self):
        return sum([len(s) for s in self.series])

    def __len__(self) -> int:
        return len(self.series)

    def __getitem__(self, index: int) -> TimeSeries:
        if self.standardize:
            return self.series[index].standardize(
                self.means[index], self.stds[index]
            )
        return self.series[index]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import argparse
import torch
import warnings
import os

warnings.filterwarnings("ignore")

from pts.model.deepvar import DeepVAREstimator
from pts import Trainer
from datasets import DATASETS, get_dataset
from pathlib import Path
import os
from utils import get_augmented_dataset

parser = argparse.ArgumentParser()

parser.add_argument(
    "outdir", type=str, help="folder to save model and training log)"
)
parser.add_argument(
    "--epochs", type=int, default=50, help="number of training epochs"
)
parser.add_argument(
    "--batch_size", type=int, default=128, help="batch size for training"
)
parser.add_argument(
    "--prediction_length", type=int, default=24, help="prediction length"
)
parser.add_argument("--context_length", type=int, default=None)
parser.add_argument(
    "--rank", type=int, default=10, help="rank of multivariate Gaussian"
)
parser.add_argument(
    "--dataset", type=str, default="electricity", choices=DATASETS
)
parser.add_argument(
    "--num_samples", type=int, default=100, help="number of sample paths"
)
parser.add_argument("--max_target_dim", type=int, default=10)
parser.add_argument(
    "--gaussian", action="store_true", help="whether perform data augmentation"
)
parser.add_argument("--num_noises", type=int, default=100)
parser.add_argument("--sigma", type=float, default=0.1)
args = parser.parse_args()


def main():
    if not os.path.exists(args.outdir):
        os.makedirs(args.outdir)

    prediction_length = args.prediction_length
    context_length = (
        4 * prediction_length
        if not args.context_length
        else args.context_length
    )
    batch_size = args.batch_size
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    ds = get_dataset(args.dataset, args.max_target_dim)
    target_dim = ds.target_dim
    freq = ds.freq
    if args.gaussian:
        train_ds = get_augmented_dataset(
            ds, num_noises=args.num_noises, sigma=args.sigma
        )
    else:
        train_ds = ds.train_ds

    input_size = 1
    if args.dataset in ["electricity", "taxi", "traffic", "solar"]:
        input_size = 47
    elif args.dataset == "wiki":
        input_size = 45
    elif args.dataset == "exchange_rate":
        input_size = 31

    trainer = Trainer(epochs=args.epochs, batch_size=batch_size, device=device)
    estimator = DeepVAREstimator(
        freq=freq,
        rank=args.rank,
        num_parallel_samples=args.num_samples,
        prediction_length=prediction_length,
        input_size=input_size,
        trainer=trainer,
        target_dim=target_dim,
        context_length=context_length,
    )

    predictor = estimator.train(training_data=train_ds)
    predictor.serialize(Path(args.outdir))


if __name__ == "__main__":
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional

import torch
import torch.nn as nn

from src.gluonts.core.component import validated
from src.gluonts.dataset.field_names import FieldName
from src.gluonts.model.predictor import Predictor
from src.gluonts.torch.model.predictor import PyTorchPredictor
from src.gluonts.torch.util import copy_parameters
from src.gluonts.transform import (
    InstanceSplitter,
    ValidationSplitSampler,
    TestSplitSampler,
    AddObservedValuesIndicator,
    Transformation,
    Chain,
    RemoveFields,
    ExpectedNumInstanceSampler,
)

from pts import Trainer
from pts.model import PyTorchEstimator
from pts.model.utils import get_module_forward_input_names

from .n_beats_network import (
    NBEATSPredictionNetwork,
    NBEATSTrainingNetwork,
    VALID_N_BEATS_STACK_TYPES,
)


class NBEATSEstimator:
    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        context_length: Optional[int] = None,
        trainer: Trainer = Trainer(),
        num_stacks: int = 30,
        widths: Optional[List[int]] = None,
        num_blocks: Optional[List[int]] = None,
        num_block_layers: Optional[List[int]] = None,
        expansion_coefficient_lengths: Optional[List[int]] = None,
        sharing: Optional[List[bool]] = None,
        stack_types: Optional[List[str]] = None,
        loss_function: Optional[str] = "MAPE",
        **kwargs,
    ) -> None:
        super().__init__(trainer=trainer, **kwargs)

        self.freq = freq
        self.prediction_length = prediction_length
        self.context_length = (
            context_length
            if context_length is not None
            else 2 * prediction_length
        )
        # num_stacks has to be handled separately because other arguments have to match its length
        self.num_stacks = num_stacks
        self.loss_function = loss_function

        self.widths = self._validate_nbeats_argument(
            argument_value=widths,
            argument_name="widths",
            default_value=[512],
            validation_condition=lambda val: val > 0,
            invalidation_message="Values of 'widths' should be > 0",
        )
        self.num_blocks = self._validate_nbeats_argument(
            argument_value=num_blocks,
            argument_name="num_blocks",
            default_value=[1],
            validation_condition=lambda val: val > 0,
            invalidation_message="Values of 'num_blocks' should be > 0",
        )
        self.num_block_layers = self._validate_nbeats_argument(
            argument_value=num_block_layers,
            argument_name="num_block_layers",
            default_value=[4],
            validation_condition=lambda val: val > 0,
            invalidation_message="Values of 'block_layers' should be > 0",
        )
        self.sharing = self._validate_nbeats_argument(
            argument_value=sharing,
            argument_name="sharing",
            default_value=[False],
            validation_condition=lambda val: True,
            invalidation_message="",
        )
        self.expansion_coefficient_lengths = self._validate_nbeats_argument(
            argument_value=expansion_coefficient_lengths,
            argument_name="expansion_coefficient_lengths",
            default_value=[32],
            validation_condition=lambda val: val > 0,
            invalidation_message="Values of 'expansion_coefficient_lengths' should be > 0",
        )
        self.stack_types = self._validate_nbeats_argument(
            argument_value=stack_types,
            argument_name="stack_types",
            default_value=["G"],
            validation_condition=lambda val: val in VALID_N_BEATS_STACK_TYPES,
            invalidation_message=f"Values of 'stack_types' should be one of {VALID_N_BEATS_STACK_TYPES}",
        )

        self.train_sampler = ExpectedNumInstanceSampler(
            num_instances=1.0, min_future=prediction_length
        )
        self.validation_sampler = ValidationSplitSampler(
            min_future=prediction_length
        )

    def _validate_nbeats_argument(
        self,
        argument_value,
        argument_name,
        default_value,
        validation_condition,
        invalidation_message,
    ):
        # set default value if applicable
        new_value = (
            argument_value if argument_value is not None else default_value
        )

        # check whether dimension of argument matches num_stack dimension
        assert len(new_value) == 1 or len(new_value) == self.num_stacks, (
            f"Invalid lengths of argument {argument_name}: {len(new_value)}. Argument must have "
            f"length 1 or {self.num_stacks} "
        )

        # check validity of actual values
        assert all(
            [validation_condition(val) for val in new_value]
        ), invalidation_message

        # make length of arguments consistent
        if len(new_value) == 1:
            return new_value * self.num_stacks
        else:
            return new_value

    # Here we do only a simple operation to convert the input data to a form
    # that can be digested by our model by only splitting the target in two, a
    # conditioning part and a to-predict part, for each training example.
    def create_transformation(self) -> Transformation:
        return Chain(
            [
                RemoveFields(
                    field_names=[
                        FieldName.FEAT_STATIC_REAL,
                        FieldName.FEAT_DYNAMIC_REAL,
                        FieldName.FEAT_DYNAMIC_CAT,
                    ]
                ),
                AddObservedValuesIndicator(
                    target_field=FieldName.TARGET,
                    output_field=FieldName.OBSERVED_VALUES,
                    dtype=self.dtype,
                ),
            ]
        )

    def create_instance_splitter(self, mode: str):
        assert mode in ["training", "validation", "test"]

        instance_sampler = {
            "training": self.train_sampler,
            "validation": self.validation_sampler,
            "test": TestSplitSampler(),
        }[mode]

        return InstanceSplitter(
            target_field=FieldName.TARGET,
            is_pad_field=FieldName.IS_PAD,
            start_field=FieldName.START,
            forecast_start_field=FieldName.FORECAST_START,
            instance_sampler=instance_sampler,
            past_length=self.context_length,
            future_length=self.prediction_length,
            time_series_fields=[FieldName.OBSERVED_VALUES],
        )

    def create_training_network(
        self, device: torch.device
    ) -> NBEATSTrainingNetwork:
        return NBEATSTrainingNetwork(
            prediction_length=self.prediction_length,
            context_length=self.context_length,
            num_stacks=self.num_stacks,
            widths=self.widths,
            num_blocks=self.num_blocks,
            num_block_layers=self.num_block_layers,
            expansion_coefficient_lengths=self.expansion_coefficient_lengths,
            sharing=self.sharing,
            stack_types=self.stack_types,
            loss_function=self.loss_function,
            freq=self.freq,
        ).to(device)

    def create_predictor(
        self,
        transformation: Transformation,
        trained_network: nn.Module,
        device: torch.device,
    ) -> Predictor:
        prediction_network = NBEATSPredictionNetwork(
            prediction_length=self.prediction_length,
            context_length=self.context_length,
            num_stacks=self.num_stacks,
            widths=self.widths,
            num_blocks=self.num_blocks,
            num_block_layers=self.num_block_layers,
            expansion_coefficient_lengths=self.expansion_coefficient_lengths,
            sharing=self.sharing,
            stack_types=self.stack_types,
        ).to(device)

        copy_parameters(trained_network, prediction_network)
        input_names = get_module_forward_input_names(prediction_network)
        prediction_splitter = self.create_instance_splitter("test")

        return PyTorchPredictor(
            input_transform=transformation + prediction_splitter,
            input_names=input_names,
            prediction_net=prediction_network,
            batch_size=self.trainer.batch_size,
            freq=self.freq,
            prediction_length=self.prediction_length,
            device=device,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Optional
from copy import copy, deepcopy

import numpy as np
import pandas as pd

from .timeseries import TimeSeries


def add_datetime_features(
    ts: TimeSeries,
    add_month_of_year: bool = False,
    add_week_of_year: bool = False,
    add_day_of_year: bool = False,
    add_day_of_month: bool = False,
    add_day_of_week: bool = False,
    add_hour_of_day: bool = False,
    add_minute_of_hour: bool = False,
    normalized: bool = True,
) -> TimeSeries:
    def _month_of_year(time_index: pd.DatetimeIndex) -> np.ndarray:
        vals = time_index.month.values - 1
        if normalized:
            vals = vals.astype(np.float32)
            vals = vals / 11.0 - 0.5
            vals = vals.reshape(-1, 1)
            return vals
        else:
            return vals.astype(np.int64)

    def _week_of_year(time_index: pd.DatetimeIndex) -> np.ndarray:
        vals = time_index.weekofyear.values - 1
        if normalized:
            vals = vals.astype(np.float32)
            vals = vals / 51.0 - 0.5
            vals = vals.reshape(-1, 1)
            return vals
        else:
            return vals.astype(np.int64)

    def _day_of_year(time_index: pd.DatetimeIndex) -> np.ndarray:
        vals = time_index.dayofyear.values - 1
        if normalized:
            vals = vals.astype(np.float32)
            vals = vals / 364.0 - 0.5
            vals = vals.reshape(-1, 1)
            return vals
        else:
            return vals.astype(np.int64)

    def _day_of_month(time_index: pd.DatetimeIndex) -> np.ndarray:
        vals = time_index.day.values - 1
        if normalized:
            vals = vals.astype(np.float32)
            vals = vals / 30.0 - 0.5
            vals = vals.reshape(-1, 1)
            return vals
        else:
            return vals.astype(np.int64)

    def _day_of_week(time_index: pd.DatetimeIndex) -> np.ndarray:
        vals = time_index.dayofweek.values
        if normalized:
            vals = vals.astype(np.float32)
            vals = vals / 6.0 - 0.5
            vals = vals.reshape(-1, 1)
            return vals
        else:
            return vals.astype(np.int64)

    def _hour_of_day(time_index: pd.DatetimeIndex) -> np.ndarray:
        vals = time_index.hour.values - 1
        if normalized:
            vals = vals.astype(np.float32)
            vals = vals / 23.0 - 0.5
            vals = vals.reshape(-1, 1)
            return vals
        else:
            return vals.astype(np.int64)

    def _minute_of_hour(time_index: pd.DatetimeIndex) -> np.ndarray:
        vals = time_index.minute.values - 1
        if normalized:
            vals = vals.astype(np.float32)
            vals = vals / 59.0 - 0.5
            vals = vals.reshape(-1, 1)
            return vals
        else:
            return vals.astype(np.int64)

    time_index = ts.time_index
    feature_type = "revealed"
    data_type = "numerical" if normalized else "categorical"
    if add_month_of_year:
        ts.add_features(
            "month_of_year",
            _month_of_year(time_index),
            data_type,
            feature_type,
        )
    if add_week_of_year:
        ts.add_features(
            "week_of_year", _week_of_year(time_index), data_type, feature_type
        )
    if add_day_of_year:
        ts.add_features(
            "day_of_year", _day_of_year(time_index), data_type, feature_type
        )
    if add_day_of_month:
        ts.add_features(
            "day_of_month", _day_of_month(time_index), data_type, feature_type
        )
    if add_day_of_week:
        ts.add_features(
            "day_of_week", _day_of_week(time_index), data_type, feature_type
        )
    if add_hour_of_day:
        ts.add_features(
            "hour_of_day", _hour_of_day(time_index), data_type, feature_type
        )
    if add_minute_of_hour:
        ts.add_features(
            "minute_of_hour",
            _minute_of_hour(time_index),
            data_type,
            feature_type,
        )
    return ts


def add_duration_feature(
    ts: TimeSeries, log_scale: Optional[float] = None, normalized: bool = True
) -> TimeSeries:
    freq = pd.infer_freq(ts.time_index)
    if freq is None:
        age = np.arange(len(ts), dtype=np.float)
    else:
        age = ts.time_index - ts.time_index[0]
        if freq == "H":
            age = age.seconds.values // 3600
        elif freq == "D":
            age = age.days.values
        else:
            raise NotImplementedError
    if log_scale is not None:
        age = np.log1p(age) / np.log(log_scale)
    if normalized:
        age = age / np.max(age) - 0.5
    ts.add_features(
        "age",
        age.reshape(-1, 1),
        data_type="numerical",
        feature_type="revealed",
    )
    return ts



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import argparse
import pickle

import numpy as np
import torch

from pts.model.simple_feedforward import SimpleFeedForwardEstimator
from pts.model.deepar import DeepAREstimator
from pts.model.lstnet import LSTNetEstimator
from pts.model.n_beats import NBEATSEstimator

from .model.lstm import LSTMEstimator
from .model.ar import AREstimator
from .trainers.SGD import SGD
from .trainers.SCott import SCott
from .trainers.Adam import Adam
from .trainers.Adagrad import Adagrad
from .trainers.SAdam import SAdam
from .trainers.SAdagrad import SAdagrad
from .trainers.SCSG import SCSG


def main():
    parser = argparse.ArgumentParser(
        description="Experiment on SCott and baselines"
    )
    # optimizer related arguments
    parser.add_argument(
        "--epochs",
        type=int,
        default=200,
        metavar="N",
        help="total number of epochs to run (default: 200)",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=1,
        metavar="N",
        help=(
            "batch size (default: 1, only for substantiating the theory, one can"
            " increase the batch size in practice)"
        ),
    )
    parser.add_argument(
        "--num_batches_per_epoch",
        type=int,
        default=50,
        metavar="N",
        help="number of batches per epoch (default: 50)",
    )
    parser.add_argument(
        "--lr", type=float, default=1e-3, help="learning rate (default: 1e-3)"
    )
    parser.add_argument(
        "--optimizer",
        type=str,
        default="SGD",
        help="optimizer used for training (default: SGD)",
    )
    parser.add_argument(
        "--nesterov",
        default=False,
        action="store_true",
        help="enable momentum for SGD (default: False)",
    )
    parser.add_argument(
        "--decreasing_step_size",
        default=False,
        action="store_true",
        help="enable decreasing_step_size (default: False)",
    )
    parser.add_argument(
        "--anchor_freq",
        type=int,
        default=20,
        help=(
            "frequency of updating the anchor gradient for control variate type"
            " optimizers (default: 20)"
        ),
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=0,
        metavar="S",
        help="random seed (default: 0)",
    )
    parser.add_argument(
        "--gamma",
        type=float,
        default=0.125,
        help="gamma variable used for the SCott",
    )

    # task (model and dataset) related arguments
    parser.add_argument(
        "--prediction_length",
        type=int,
        default=1,
        metavar="N",
        help="prediction length (dimension of the prediction space, default: 1)",
    )
    parser.add_argument(
        "--context_length",
        type=int,
        default=8,
        metavar="N",
        help="context length (dimension of the context space, default: 8)",
    )
    parser.add_argument(
        "--input_size",
        type=int,
        default=1,
        metavar="N",
        help="input size of DeepAR (default: 1)",
    )
    parser.add_argument(
        "--freq",
        type=str,
        default="1H",
        help="frequency of the time series data (default: 1 hour)",
    )
    parser.add_argument(
        "--task_name",
        type=str,
        default="exchange_rate",
        help="name of the task to run (default: exchange rate)",
    )
    parser.add_argument(
        "--tensorboard_path",
        type=str,
        default="./runs/test",
        help="path to log the results from trainer",
    )
    parser.add_argument(
        "--dataset",
        type=str,
        default="exchange_rate",
        help="dataset to use in the experiment (default: exchange_rate)",
    )
    parser.add_argument(
        "--num_duplicates",
        type=int,
        default=1,
        help="number of duplicates to use (for synthetic data experiment)",
    )
    parser.add_argument(
        "--num_strata",
        type=int,
        default=8,
        help="number of strata in the dataset (default: 8)",
    )
    parser.add_argument(
        "--data_file",
        type=str,
        default="None",
        help="name of the file dataset to use",
    )

    parser.add_argument(
        "--eval_model",
        default=True,
        action="store_true",
        help="enable validation during training (default: True)",
    )

    parser.add_argument(
        "--model", type=str, default="mlp", help="model to use (default: MLP)"
    )
    parser.add_argument(
        "--hidden_layer_size",
        type=int,
        default=100,
        metavar="S",
        help="hidden layer size for MLP model (default: 100)",
    )

    args = parser.parse_args()

    torch.manual_seed(args.seed)
    np.random.seed(args.seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    num_hidden_dimensions = None

    # get dataset
    with open("./dataset/" + args.data_file + ".csv", "rb") as input:
        data_package = pickle.load(input)
        freq = args.freq
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # get trainer
    if args.optimizer == "SGD":
        trainer = SGD(
            epochs=args.epochs,
            task_name=args.task_name,
            num_batches_per_epoch=args.num_batches_per_epoch,
            batch_size=args.batch_size,
            learning_rate=args.lr,
            decreasing_step_size=args.decreasing_step_size,
            nesterov=args.nesterov,
            eval_model=args.eval_model,
            tensorboard_path=args.tensorboard_path,
            device=device,
        )

    if args.optimizer == "Adam":
        trainer = Adam(
            epochs=args.epochs,
            task_name=args.task_name,
            num_batches_per_epoch=args.num_batches_per_epoch,
            batch_size=args.batch_size,
            learning_rate=args.lr,
            decreasing_step_size=args.decreasing_step_size,
            eval_model=args.eval_model,
            tensorboard_path=args.tensorboard_path,
            device=device,
        )

    if args.optimizer == "Adagrad":
        trainer = Adagrad(
            epochs=args.epochs,
            task_name=args.task_name,
            num_batches_per_epoch=args.num_batches_per_epoch,
            batch_size=args.batch_size,
            learning_rate=args.lr,
            decreasing_step_size=args.decreasing_step_size,
            eval_model=args.eval_model,
            tensorboard_path=args.tensorboard_path,
            device=device,
        )

    if args.optimizer == "SCott":
        trainer = SCott(
            epochs=args.epochs,
            task_name=args.task_name,
            num_batches_per_epoch=args.num_batches_per_epoch,
            batch_size=args.batch_size,
            num_strata=args.num_strata,
            learning_rate=args.lr,
            freq=args.anchor_freq,
            decreasing_step_size=args.decreasing_step_size,
            nesterov=args.nesterov,
            eval_model=args.eval_model,
            tensorboard_path=args.tensorboard_path,
            gamma=args.gamma,
            device=device,
        )

    if args.optimizer == "SAdam":
        trainer = SAdam(
            epochs=args.epochs,
            task_name=args.task_name,
            num_batches_per_epoch=args.num_batches_per_epoch,
            batch_size=args.batch_size,
            num_strata=args.num_strata,
            learning_rate=args.lr,
            freq=args.anchor_freq,
            decreasing_step_size=args.decreasing_step_size,
            eval_model=args.eval_model,
            tensorboard_path=args.tensorboard_path,
            gamma=args.gamma,
            device=device,
        )

    if args.optimizer == "SAdagrad":
        trainer = SAdagrad(
            epochs=args.epochs,
            task_name=args.task_name,
            num_batches_per_epoch=args.num_batches_per_epoch,
            batch_size=args.batch_size,
            num_strata=args.num_strata,
            learning_rate=args.lr,
            freq=args.anchor_freq,
            decreasing_step_size=args.decreasing_step_size,
            eval_model=args.eval_model,
            tensorboard_path=args.tensorboard_path,
            gamma=args.gamma,
            device=device,
        )

    if args.optimizer == "SCSG":
        trainer = SCSG(
            epochs=args.epochs,
            task_name=args.task_name,
            num_batches_per_epoch=args.num_batches_per_epoch,
            batch_size=args.batch_size,
            num_strata=args.num_strata,
            learning_rate=args.lr,
            freq=args.anchor_freq,
            decreasing_step_size=args.decreasing_step_size,
            nesterov=args.nesterov,
            eval_model=args.eval_model,
            tensorboard_path=args.tensorboard_path,
            device=device,
        )

    # get model
    if args.model == "mlp":
        estimator = SimpleFeedForwardEstimator(
            freq=freq,
            prediction_length=args.prediction_length,
            context_length=args.context_length,
            num_hidden_dimensions=num_hidden_dimensions,
            trainer=trainer,
        )

    if args.model == "lstm":
        estimator = LSTMEstimator(
            freq=freq,
            prediction_length=args.prediction_length,
            context_length=args.context_length,
            input_size=args.input_size,
            hidden_layer_size=args.hidden_layer_size,
            trainer=trainer,
        )

    if args.model == "linear":
        estimator = AREstimator(
            freq=freq,
            prediction_length=args.prediction_length,
            context_length=args.context_length,
            trainer=trainer,
        )

    if args.model == "deepar":
        estimator = DeepAREstimator(
            freq=freq,
            prediction_length=args.prediction_length,
            context_length=args.context_length,
            input_size=42,
            cardinality=[70],
            use_feat_static_cat=True,
            trainer=trainer,
        )

    if args.model == "nbeats":
        estimator = NBEATSEstimator(
            freq=freq,
            prediction_length=args.prediction_length,
            context_length=args.context_length,
            trainer=trainer,
        )

    if args.model == "lstnet":
        estimator = LSTNetEstimator(
            skip_size=1,
            ar_window=1,
            num_series=1,
            channels=6,
            kernel_size=2,
            context_length=args.context_length,
            freq=freq,
            prediction_length=args.prediction_length,
            trainer=trainer,
        )

    # get proper training entry
    if args.optimizer in ["SGD", "Adam", "Adagrad", "SCSG", "BatchSGD"]:
        predictor = estimator.train(data_package=data_package)
    else:
        predictor = estimator.stratified_train(data_package=data_package)


if __name__ == "__main__":
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional

import torch

from src.gluonts.core.component import validated
from src.gluonts.dataset.field_names import FieldName
from src.gluonts.time_feature import TimeFeature
from src.gluonts.torch.util import copy_parameters
from src.gluonts.torch.model.predictor import PyTorchPredictor
from src.gluonts.model.predictor import Predictor
from src.gluonts.transform import (
    Transformation,
    Chain,
    InstanceSplitter,
    ValidationSplitSampler,
    TestSplitSampler,
    ExpectedNumInstanceSampler,
    RemoveFields,
    RenameFields,
    AsNumpyArray,
    ExpandDimArray,
    AddObservedValuesIndicator,
    AddTimeFeatures,
    VstackFeatures,
    SetFieldIfNotPresent,
    TargetDimIndicator,
)

from pts import Trainer
from pts.model import PyTorchEstimator
from pts.model.utils import get_module_forward_input_names
from pts.feature import (
    fourier_time_features_from_frequency,
    lags_for_fourier_time_features_from_frequency,
)

from .transformer_tempflow_network import (
    TransformerTempFlowTrainingNetwork,
    TransformerTempFlowPredictionNetwork,
)


class TransformerTempFlowEstimator(PyTorchEstimator):
    @validated()
    def __init__(
        self,
        input_size: int,
        freq: str,
        prediction_length: int,
        target_dim: int,
        trainer: Trainer = Trainer(),
        context_length: Optional[int] = None,
        d_model: int = 32,
        dim_feedforward_scale: int = 4,
        act_type: str = "gelu",
        num_heads: int = 8,
        num_encoder_layers: int = 3,
        num_decoder_layers: int = 3,
        num_parallel_samples: int = 100,
        dropout_rate: float = 0.1,
        use_feat_dynamic_real: bool = False,
        flow_type="RealNVP",
        n_blocks=3,
        hidden_size=100,
        n_hidden=2,
        conditioning_length: int = 200,
        dequantize: bool = False,
        scaling: bool = True,
        pick_incomplete: bool = False,
        lags_seq: Optional[List[int]] = None,
        time_features: Optional[List[TimeFeature]] = None,
        **kwargs,
    ) -> None:
        super().__init__(trainer=trainer, **kwargs)

        self.freq = freq
        self.context_length = (
            context_length if context_length is not None else prediction_length
        )

        self.input_size = input_size
        self.prediction_length = prediction_length
        self.target_dim = target_dim

        self.d_model = d_model
        self.num_heads = num_heads
        self.act_type = act_type
        self.dim_feedforward_scale = dim_feedforward_scale
        self.num_encoder_layers = num_encoder_layers
        self.num_decoder_layers = num_decoder_layers

        self.num_parallel_samples = num_parallel_samples
        self.dropout_rate = dropout_rate

        self.use_feat_dynamic_real = use_feat_dynamic_real

        self.flow_type = flow_type
        self.n_blocks = n_blocks
        self.hidden_size = hidden_size
        self.n_hidden = n_hidden
        self.conditioning_length = conditioning_length
        self.dequantize = dequantize

        self.lags_seq = (
            lags_seq
            if lags_seq is not None
            else lags_for_fourier_time_features_from_frequency(freq_str=freq)
        )

        self.time_features = (
            time_features
            if time_features is not None
            else fourier_time_features_from_frequency(self.freq)
        )

        self.history_length = self.context_length + max(self.lags_seq)
        self.pick_incomplete = pick_incomplete
        self.scaling = scaling

        self.train_sampler = ExpectedNumInstanceSampler(
            num_instances=1.0,
            min_past=0 if pick_incomplete else self.history_length,
            min_future=prediction_length,
        )

        self.validation_sampler = ValidationSplitSampler(
            min_past=0 if pick_incomplete else self.history_length,
            min_future=prediction_length,
        )

    def create_transformation(self) -> Transformation:
        remove_field_names = [FieldName.FEAT_DYNAMIC_CAT]
        if not self.use_feat_dynamic_real:
            remove_field_names.append(FieldName.FEAT_DYNAMIC_REAL)

        return Chain(
            [
                RemoveFields(field_names=remove_field_names),
                AsNumpyArray(
                    field=FieldName.TARGET,
                    expected_ndim=2,
                ),
                # maps the target to (1, T)
                # if the target data is uni dimensional
                ExpandDimArray(
                    field=FieldName.TARGET,
                    axis=None,
                ),
                AddObservedValuesIndicator(
                    target_field=FieldName.TARGET,
                    output_field=FieldName.OBSERVED_VALUES,
                ),
                AddTimeFeatures(
                    start_field=FieldName.START,
                    target_field=FieldName.TARGET,
                    output_field=FieldName.FEAT_TIME,
                    time_features=self.time_features,
                    pred_length=self.prediction_length,
                ),
                VstackFeatures(
                    output_field=FieldName.FEAT_TIME,
                    input_fields=[FieldName.FEAT_TIME]
                    + (
                        [FieldName.FEAT_DYNAMIC_REAL]
                        if self.use_feat_dynamic_real
                        else []
                    ),
                ),
                SetFieldIfNotPresent(
                    field=FieldName.FEAT_STATIC_CAT, value=[0]
                ),
                TargetDimIndicator(
                    field_name="target_dimension_indicator",
                    target_field=FieldName.TARGET,
                ),
                AsNumpyArray(field=FieldName.FEAT_STATIC_CAT, expected_ndim=1),
            ]
        )

    def create_instance_splitter(self, mode: str):
        assert mode in ["training", "validation", "test"]

        instance_sampler = {
            "training": self.train_sampler,
            "validation": self.validation_sampler,
            "test": TestSplitSampler(),
        }[mode]

        return InstanceSplitter(
            target_field=FieldName.TARGET,
            is_pad_field=FieldName.IS_PAD,
            start_field=FieldName.START,
            forecast_start_field=FieldName.FORECAST_START,
            instance_sampler=instance_sampler,
            past_length=self.history_length,
            future_length=self.prediction_length,
            time_series_fields=[
                FieldName.FEAT_TIME,
                FieldName.OBSERVED_VALUES,
            ],
        ) + (
            RenameFields(
                {
                    f"past_{FieldName.TARGET}": f"past_{FieldName.TARGET}_cdf",
                    f"future_{FieldName.TARGET}": f"future_{FieldName.TARGET}_cdf",
                }
            )
        )

    def create_training_network(
        self, device: torch.device
    ) -> TransformerTempFlowTrainingNetwork:
        return TransformerTempFlowTrainingNetwork(
            input_size=self.input_size,
            target_dim=self.target_dim,
            num_heads=self.num_heads,
            act_type=self.act_type,
            d_model=self.d_model,
            dim_feedforward_scale=self.dim_feedforward_scale,
            num_encoder_layers=self.num_encoder_layers,
            num_decoder_layers=self.num_decoder_layers,
            history_length=self.history_length,
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            dropout_rate=self.dropout_rate,
            lags_seq=self.lags_seq,
            scaling=self.scaling,
            flow_type=self.flow_type,
            n_blocks=self.n_blocks,
            hidden_size=self.hidden_size,
            n_hidden=self.n_hidden,
            conditioning_length=self.conditioning_length,
            dequantize=self.dequantize,
        ).to(device)

    def create_predictor(
        self,
        transformation: Transformation,
        trained_network: TransformerTempFlowTrainingNetwork,
        device: torch.device,
    ) -> Predictor:
        prediction_network = TransformerTempFlowPredictionNetwork(
            input_size=self.input_size,
            target_dim=self.target_dim,
            num_heads=self.num_heads,
            act_type=self.act_type,
            d_model=self.d_model,
            dim_feedforward_scale=self.dim_feedforward_scale,
            num_encoder_layers=self.num_encoder_layers,
            num_decoder_layers=self.num_decoder_layers,
            history_length=self.history_length,
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            dropout_rate=self.dropout_rate,
            lags_seq=self.lags_seq,
            scaling=self.scaling,
            flow_type=self.flow_type,
            n_blocks=self.n_blocks,
            hidden_size=self.hidden_size,
            n_hidden=self.n_hidden,
            conditioning_length=self.conditioning_length,
            dequantize=self.dequantize,
            num_parallel_samples=self.num_parallel_samples,
        ).to(device)

        copy_parameters(trained_network, prediction_network)
        input_names = get_module_forward_input_names(prediction_network)
        prediction_splitter = self.create_instance_splitter("test")

        return PyTorchPredictor(
            input_transform=transformation + prediction_splitter,
            input_names=input_names,
            prediction_net=prediction_network,
            batch_size=self.trainer.batch_size,
            freq=self.freq,
            prediction_length=self.prediction_length,
            device=device,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pts.dataset.repository.datasets import get_dataset, dataset_recipes
from pts.dataset.utils import to_pandas
import os
import numpy as np
import pandas as pd
from pts.dataset import ListDataset
import torch
import pickle
import json
import random
import os


def _get_mixed_pattern(x, pattern):
    if pattern == "sin":
        return (torch.sin(x) + torch.normal(0, 0.1, size=x.shape)) * 5
    elif pattern == "linear":
        return x * 2 + torch.normal(0, 1, size=x.shape)
    elif pattern == "quadratic":
        return x**1.5 + torch.normal(0, 5, size=x.shape)
    else:
        return torch.sqrt(x) + torch.normal(0, 1, size=x.shape)


def get_mixed_pattern(unit_length=16, num_duplicates=1000):
    freq = "1H"
    context_length = 3 * unit_length
    prediction_length = unit_length
    len_sample = context_length + prediction_length
    dataset_group = [[] for j in range(16)]
    whole_data = []
    val_data = []
    ret = dict()
    start = pd.Timestamp("01-01-2000", freq=freq)
    patterns = [
        ["sin", "linear", "quadratic", "sqrt"],
        ["sqrt", "quadratic", "linear", "sin"],
        ["linear", "sqrt", "sin", "quadratic"],
        ["quadratic", "sin", "sqrt", "linear"],
    ]
    pattern_number = 4
    for m, pattern in enumerate(patterns):
        for gid in range(pattern_number):
            for j in range(num_duplicates):
                context = torch.arange(context_length, dtype=torch.float)
                for i in range(1, pattern_number):
                    context[unit_length * (i - 1) : unit_length * i] = (
                        _get_mixed_pattern(
                            context[unit_length * (i - 1) : unit_length * i]
                            - unit_length * (i - 1),
                            pattern[(gid + i) % pattern_number],
                        )
                    )
                ts_sample = torch.cat(
                    [
                        context,
                        _get_mixed_pattern(
                            torch.arange(prediction_length, dtype=torch.float),
                            pattern[gid],
                        ),
                    ]
                )
                whole_data.append({"target": ts_sample, "start": start})
                if j % 5 == 0:
                    val_data.append(
                        {
                            "target": ts_sample
                            + torch.normal(0, 1, ts_sample.shape),
                            "start": start,
                        }
                    )
                dataset_group[m * 4 + gid].append(
                    {"target": ts_sample, "start": start}
                )
    print(
        "Generating the synthetic training data, the total number of training"
        " examples:",
        len(whole_data),
    )
    ret["group_ratio"] = [len(i) / len(whole_data) for i in dataset_group]
    random.shuffle(whole_data)
    group_data = []
    ret["whole_data"] = ListDataset(whole_data, freq=freq)
    ret["val_data"] = ListDataset(val_data, freq=freq)
    for group in dataset_group:
        random.shuffle(group)
        group_data.append(ListDataset(group, freq=freq))
    ret["group_data"] = group_data
    # save to files
    os.makedirs("./dataset", exist_ok=True)
    with open("./dataset/synthetic.csv", "wb") as output:
        pickle.dump(ret, output)
    print("Finished the pre-processing of synthetic dataset")

    return True


def group_electricity_cv(
    num_ts=10,
    num_groups=14,
    context_length=72,
    prediction_length=12,
    file_name="default",
):
    dataset = get_dataset("electricity", regenerate=True)
    len_sample = context_length + prediction_length
    dataset_group = [[] for i in range(num_groups)]
    train_full_data = []
    test_full_data = []
    ret = dict()
    train_it = iter(dataset.train)
    test_it = iter(dataset.test)
    date_checkpoint = [
        "2012-03-01",
        "2012-06-01",
        "2012-09-01",
        "2012-12-01",
        "2013-03-01",
        "2013-06-01",
        "2013-09-01",
        "2013-12-01",
        "2014-03-01",
    ]
    # get ready the training data
    for i in range(num_ts):
        train_entry = next(train_it)
        unsplit_ts = train_entry["target"]
        unsplit_start = train_entry["start"]
        t = unsplit_start
        start_date = 4

        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            for j, date_ckpt in enumerate(date_checkpoint):
                if unsplit_start < pd.Timestamp(date_ckpt):
                    sid = j
                    break
                elif unsplit_start > pd.Timestamp(date_checkpoint[-1]):
                    sid = len(date_checkpoint)
                    break
            gid = ((start_date + 1) % 7) + sid * 7
            start_date += 1
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            train_full_data.append(
                {
                    "target": ts_slice,
                    "start": t,
                    "feat_static_cat": np.array([gid]),
                }
            )
            dataset_group[gid].append(
                {
                    "target": ts_slice,
                    "start": t,
                    "feat_static_cat": np.array([gid]),
                }
            )
            unsplit_start += pd.Timedelta(hours=prediction_length)

    # get ready the test data
    for i in range(int(num_ts * 0.2)):
        test_entry = next(test_it)
        unsplit_ts = test_entry["target"]
        unsplit_start = test_entry["start"]
        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            test_full_data.append(
                {
                    "target": ts_slice,
                    "start": unsplit_start,
                    "feat_static_cat": test_entry["feat_static_cat"],
                }
            )

    print(
        "Generating the electricity training data, the total number of"
        " training examples:",
        len(train_full_data),
    )
    ret["group_ratio"] = [len(i) / len(train_full_data) for i in dataset_group]
    random.shuffle(train_full_data)
    ret["whole_data"] = ListDataset(
        train_full_data, freq=dataset.metadata.freq
    )
    random.shuffle(test_full_data)
    ret["val_data"] = ListDataset(test_full_data, freq=dataset.metadata.freq)
    group_data_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_data_list.append(ListDataset(group, freq=dataset.metadata.freq))
    ret["group_data"] = group_data_list
    os.makedirs("./dataset", exist_ok=True)
    with open("./dataset/" + file_name + ".csv", "wb") as output:
        pickle.dump(ret, output)
    print("Finished pre-processing of the electricity dataset")
    return True

    dataset = get_dataset("traffic")
    len_sample = context_length + prediction_length
    dataset_group = [[] for i in range(num_groups)]
    train_full_data = []
    test_full_data = []
    ret = dict()
    train_it = iter(dataset.train)
    test_it = iter(dataset.test)
    # num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    date_checkpoint = ["2016-01-01"]
    # get ready the training data
    for i in range(num_ts):
        train_entry = next(train_it)
        unsplit_ts = train_entry["target"]
        unsplit_start = train_entry["start"]
        t = unsplit_start
        start_date = 4

        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            for j, date_ckpt in enumerate(date_checkpoint):
                if unsplit_start < pd.Timestamp(date_ckpt):
                    sid = j
                    break
                elif unsplit_start > pd.Timestamp(date_checkpoint[-1]):
                    sid = len(date_checkpoint)
                    break
            gid = ((start_date + 1) % 7) + sid * 7
            start_date += 1
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            train_full_data.append(
                {
                    "target": ts_slice,
                    "start": t,
                    "feat_static_cat": train_entry["feat_static_cat"],
                }
            )
            dataset_group[gid].append(
                {
                    "target": ts_slice,
                    "start": t,
                    "feat_static_cat": train_entry["feat_static_cat"],
                }
            )
            unsplit_start += pd.Timedelta(hours=prediction_length)

    # get ready the test data
    for i in range(int(num_ts * 0.2)):
        test_entry = next(test_it)
        unsplit_ts = test_entry["target"]
        unsplit_start = test_entry["start"]
        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            test_full_data.append(
                {
                    "target": ts_slice,
                    "start": unsplit_start,
                    "feat_static_cat": test_entry["feat_static_cat"],
                }
            )

    print("total number of training examples: ", len(train_full_data))
    ret["group_ratio"] = [len(i) / len(train_full_data) for i in dataset_group]
    print("ratio for each group: ", ret["group_ratio"])
    random.shuffle(train_full_data)
    ret["whole_data"] = ListDataset(
        train_full_data, freq=dataset.metadata.freq
    )
    random.shuffle(test_full_data)
    ret["val_data"] = ListDataset(test_full_data, freq=dataset.metadata.freq)
    group_data_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_data_list.append(ListDataset(group, freq=dataset.metadata.freq))
    ret["group_data"] = group_data_list
    os.makedirs("./dataset", exist_ok=True)
    with open("./dataset/" + file_name + ".csv", "wb") as output:
        pickle.dump(ret, output)
    return True


def group_exchangerate_cv(
    num_ts=10,
    num_groups=14,
    context_length=15,
    prediction_length=10,
    file_name="default",
):
    dataset = get_dataset("exchange_rate", regenerate=True)
    len_sample = context_length + prediction_length
    dataset_group = [[] for i in range(num_groups)]
    train_full_data = []
    test_full_data = []
    ret = dict()
    train_it = iter(dataset.train)
    test_it = iter(dataset.test)
    # num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    date_checkpoint = ["1994-01-01", "1998-01-01", "2002-01-01"]
    for i in range(num_ts):
        train_entry = next(train_it)
        unsplit_ts = train_entry["target"]
        unsplit_start = train_entry["start"]
        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            for j, date_ckpt in enumerate(date_checkpoint):
                if unsplit_start < pd.Timestamp(date_ckpt):
                    sid = j
                    break
                elif unsplit_start > pd.Timestamp(date_checkpoint[-1]):
                    sid = len(date_checkpoint)
                    break
            gid = i * 4 + sid
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            train_full_data.append(
                {
                    "target": ts_slice,
                    "start": unsplit_start,
                    "feat_static_cat": train_entry["feat_static_cat"],
                }
            )
            dataset_group[gid].append(
                {
                    "target": ts_slice,
                    "start": unsplit_start,
                    "feat_static_cat": train_entry["feat_static_cat"],
                }
            )
            unsplit_start += pd.Timedelta("1D") * prediction_length
    # get ready the test data
    for i in range(int(num_ts * 0.2)):
        test_entry = next(test_it)
        unsplit_ts = test_entry["target"]
        unsplit_start = test_entry["start"]
        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            test_full_data.append(
                {
                    "target": ts_slice,
                    "start": unsplit_start,
                    "feat_static_cat": test_entry["feat_static_cat"],
                }
            )
    print(
        "Generating the exchange rate training data, the total number of"
        " training examples:",
        len(train_full_data),
    )
    ret["group_ratio"] = [len(i) / len(train_full_data) for i in dataset_group]
    random.shuffle(train_full_data)
    ret["whole_data"] = ListDataset(
        train_full_data, freq=dataset.metadata.freq
    )
    random.shuffle(test_full_data)
    ret["val_data"] = ListDataset(test_full_data, freq=dataset.metadata.freq)
    group_data_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_data_list.append(ListDataset(group, freq=dataset.metadata.freq))
    ret["group_data"] = group_data_list
    os.makedirs("./dataset", exist_ok=True)
    with open("./dataset/" + file_name + ".csv", "wb") as output:
        pickle.dump(ret, output)
    print("Finished pre-processing the exchange rate dataset")
    return True


def group_traffic_cv(
    num_ts=10,
    num_groups=14,
    context_length=72,
    prediction_length=12,
    file_name="default",
):
    dataset = get_dataset("traffic", regenerate=True)
    len_sample = context_length + prediction_length
    dataset_group = [[] for i in range(num_groups)]
    train_full_data = []
    test_full_data = []
    ret = dict()
    train_it = iter(dataset.train)
    test_it = iter(dataset.test)
    # num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    date_checkpoint = [
        "2015-03-01",
        "2015-06-01",
        "2015-09-01",
        "2015-12-01",
        "2016-03-01",
        "2016-06-01",
    ]
    # get ready the training data
    for i in range(num_ts):
        train_entry = next(train_it)
        unsplit_ts = train_entry["target"]
        unsplit_start = train_entry["start"]
        t = unsplit_start
        start_date = 4
        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            for j, date_ckpt in enumerate(date_checkpoint):
                if unsplit_start < pd.Timestamp(date_ckpt):
                    sid = j
                    break
                elif unsplit_start > pd.Timestamp(date_checkpoint[-1]):
                    sid = len(date_checkpoint)
                    break
            gid = ((start_date + 1) % 7) + sid * 7
            start_date += 1
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            train_full_data.append(
                {
                    "target": ts_slice,
                    "start": t,
                    "feat_static_cat": train_entry["feat_static_cat"],
                }
            )
            dataset_group[gid].append(
                {
                    "target": ts_slice,
                    "start": t,
                    "feat_static_cat": train_entry["feat_static_cat"],
                }
            )
            unsplit_start += pd.Timedelta(hours=prediction_length)

    # get ready the test data
    for i in range(int(num_ts * 0.2)):
        test_entry = next(test_it)
        unsplit_ts = test_entry["target"]
        unsplit_start = test_entry["start"]
        for ts_sample_start in range(
            0, len(unsplit_ts) - len_sample, prediction_length
        ):
            ts_slice = unsplit_ts[
                ts_sample_start : ts_sample_start + len_sample
            ]
            test_full_data.append(
                {
                    "target": ts_slice,
                    "start": unsplit_start,
                    "feat_static_cat": test_entry["feat_static_cat"],
                }
            )

    print(
        "Generating the traffic training data, the total number of training"
        " examples:",
        len(train_full_data),
    )
    ret["group_ratio"] = [len(i) / len(train_full_data) for i in dataset_group]
    random.shuffle(train_full_data)
    ret["whole_data"] = ListDataset(
        train_full_data, freq=dataset.metadata.freq
    )
    random.shuffle(test_full_data)
    ret["val_data"] = ListDataset(test_full_data, freq=dataset.metadata.freq)
    group_data_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_data_list.append(ListDataset(group, freq=dataset.metadata.freq))
    ret["group_data"] = group_data_list
    os.makedirs("./dataset", exist_ok=True)
    with open("./dataset/" + file_name + ".csv", "wb") as output:
        pickle.dump(ret, output)
    print("Finished pre-processing the traffic dataset")
    return True


def main():
    get_mixed_pattern(unit_length=24, num_duplicates=2000)
    group_traffic_cv(
        num_ts=800,
        num_groups=49,
        context_length=72,
        prediction_length=24,
        file_name="traffic",
    )
    group_exchangerate_cv(
        num_ts=8,
        num_groups=32,
        context_length=8,
        prediction_length=1,
        file_name="exchange_rate",
    )
    group_electricity_cv(
        num_ts=300,
        num_groups=70,
        context_length=72,
        prediction_length=24,
        file_name="electricity",
    )
    print(
        "Finished the preprocessing data, please verify ./dataset/ contains four"
        " .csv files"
    )


if __name__ == "__main__":
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Optional, List, Tuple, Union

import numpy as np
import torch as pt
from torch import Tensor, BoolTensor, nn
from torch.nn import init, functional as F

from tslib.nn.transformer import PositionwiseFFN
from .kernel import AttentionKernel
from .disc import SimpleDiscriminator


class EncoderModule(nn.Module):
    def __init__(
        self,
        d_data: int,
        d_feats: int,
        d_hidden: int,
        d_value: int,
        *window_size: int,
        tie_casts: bool = False,
    ) -> None:
        super(EncoderModule, self).__init__()
        self.d_data = d_data
        self.d_feats = d_feats
        self.d_hidden = d_hidden
        self.window_size = window_size
        self.tie_casts = tie_casts
        assert len(self.window_size) <= self.d_hidden

        d_shape = [d_hidden // len(window_size)] * len(window_size)
        d_shape[-1] += self.d_hidden % len(window_size)
        self._shape_weights = nn.ParameterList()
        self._shape_biases = nn.ParameterList()
        for d_s, wsize in zip(d_shape, window_size):
            self._shape_weights.append(
                nn.Parameter(Tensor(d_s, d_data + d_feats, wsize))
            )
            self._shape_biases.append(nn.Parameter(Tensor(d_s)))

        self._inter_weight = nn.Parameter(Tensor(d_value, d_data + d_feats))
        self._inter_bias = nn.Parameter(Tensor(d_value))
        if self.tie_casts:
            self._extra_weight = self._inter_weight
            self._extra_bias = self._inter_bias
        else:
            self._extra_weight = nn.Parameter(
                Tensor(d_value, d_data + d_feats)
            )
            self._extra_bias = nn.Parameter(Tensor(d_value))

        self._reset_parameters()

    def _reset_parameters(self):
        for _shape_weight in self._shape_weights:
            init.xavier_uniform_(_shape_weight)
        for _shape_bias in self._shape_biases:
            init.zeros_(_shape_bias)
        init.xavier_uniform_(self._inter_weight)
        init.zeros_(self._inter_bias)
        if not self.tie_casts:
            init.xavier_uniform_(self._extra_weight)
            init.zeros_(self._extra_bias)

    def _compute_shape(self, data: Tensor):
        s = data.transpose(1, 2)
        shape = []
        for i, wsize in enumerate(self.window_size):
            # left_pad + right_pad = window_size - 1
            left_pad = wsize // 2
            right_pad = wsize // 2 - (1 - wsize % 2)
            x = F.pad(
                s,
                [left_pad, right_pad],
                mode="constant",
                value=0.0,
            )
            x = F.conv1d(x, self._shape_weights[i], self._shape_biases[i])
            shape.append(x)
        shape = pt.cat(shape, dim=1)
        shape = shape.transpose(1, 2)
        return shape

    def _compute_value(self, data: Tensor):
        inter_value = F.linear(data, self._inter_weight, self._inter_bias)
        extra_value = F.linear(
            data[:, max(self.window_size) :],
            self._extra_weight,
            self._extra_bias,
        )
        return inter_value, extra_value

    def forward(
        self,
        data: Tensor,
        feats: Optional[Tensor],
    ):
        if feats is None:
            assert self.d_feats == 0
        else:
            assert self.d_feats == feats.size(2)
            data = pt.cat([data, feats], dim=2)
        shape = self._compute_shape(data)
        inter_value, extra_value = self._compute_value(data)
        return shape, inter_value, extra_value


class DecoderModule(nn.Module):
    def __init__(
        self,
        d_data: int,
        d_hidden: int,
    ) -> None:
        super(DecoderModule, self).__init__()
        self.d_data = d_data
        self.d_hidden = d_hidden

        self.weight = nn.Parameter(Tensor(self.d_data, self.d_hidden))
        self.bias = nn.Parameter(Tensor(d_data))

        self._reset_parameters()

    def _reset_parameters(self):
        init.xavier_uniform_(self.weight)
        init.zeros_(self.bias)

    def forward(self, data: Tensor):
        return F.linear(data, self.weight, self.bias)


class AttentionBlock(nn.Module):
    def __init__(
        self,
        encoder: EncoderModule,
        kernel: AttentionKernel,
        decoder: DecoderModule,
    ) -> None:
        super(AttentionBlock, self).__init__()
        self.encoder = encoder
        self.kernel = kernel
        self.decoder = decoder
        self.window_size = max(self.encoder.window_size)

        self.register_buffer("shape", None, persistent=False)
        self.register_buffer("query", None, persistent=False)
        self.register_buffer("key", None, persistent=False)
        self.register_buffer("inter_score", None, persistent=False)
        self.register_buffer("extra_score", None, persistent=False)
        self.register_buffer("inter_value", None, persistent=False)
        self.register_buffer("extra_value", None, persistent=False)

    def forward(
        self,
        data: Tensor,
        feats: Optional[Tensor],
        mask: Optional[BoolTensor],
    ) -> Tuple[Tensor, Tensor]:
        shape, inter_value, extra_value = self.encoder(data, feats)
        self.shape = shape.detach()
        self.inter_value = inter_value.detach()
        self.extra_value = extra_value.detach()
        interp, extrap = self.kernel(shape, inter_value, extra_value, mask)
        self.query = self.kernel._query
        self.key = self.kernel._key
        self.inter_score = self.kernel._inter_score
        self.extra_score = self.kernel._extra_score
        interp = self.decoder(interp)
        extrap = self.decoder(extrap)
        return interp, extrap


class AdversarialBlock(AttentionBlock):
    def __init__(
        self,
        encoder: EncoderModule,
        kernel: AttentionKernel,
        decoder: DecoderModule,
        disc: SimpleDiscriminator,
    ) -> None:
        super(AdversarialBlock, self).__init__(encoder, kernel, decoder)
        self.disc = disc
        self.generative()
        self.register_buffer("prob_domain", None, persistent=False)

    def generative(self) -> None:
        self._generative = True
        self.disc.requires_grad_(False)
        self.encoder.requires_grad_(True)
        self.decoder.requires_grad_(True)
        self.kernel.requires_grad_(True)

    def discriminative(self) -> None:
        self._generative = False
        self.disc.requires_grad_(True)
        self.encoder.requires_grad_(False)
        self.decoder.requires_grad_(False)
        self.kernel.requires_grad_(False)

    def forward(
        self,
        data: Tensor,
        feats: Optional[Tensor],
        mask: Optional[BoolTensor],
    ):
        shape, inter_value, extra_value = self.encoder(data, feats)
        self.shape = shape.detach()
        self.inter_value = inter_value.detach()
        self.extra_value = extra_value.detach()
        interp, extrap = self.kernel(shape, inter_value, extra_value, mask)
        self.query = self.kernel._query.detach()
        self.key = self.kernel._key.detach()
        self.inter_score = self.kernel._inter_score
        self.extra_score = self.kernel._extra_score

        # TODO: disc both queries and keys
        features = self.kernel._query
        if not self._generative:
            features = features.detach()
        self.prob_domain = self.disc(features)

        interp = self.decoder(interp)
        extrap = self.decoder(extrap)
        return interp, extrap



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple
import numpy as np
import pandas as pd
from tqdm.auto import tqdm
from tsbench.analysis.utils import (
    loocv_split,
    num_fitting_processes,
    run_parallel,
)
from tsbench.config import Config, ModelConfig
from tsbench.evaluations.metrics import Performance
from tsbench.evaluations.tracking import ModelTracker
from tsbench.surrogate import AutoGluonSurrogate, Surrogate
from .metrics import mrr, ndcg, nrmse, precision_k, smape


class SurrogateAnalyzer:
    """
    The surrogate analyzer evaluates the performance of a surrogate model with
    respect to ranking and regression metrics.

    The analysis is run in parallel and should, thus, not be used in a Jupyter
    notebook. Instead, consider using the `tsbench` CLI.
    """

    def __init__(
        self,
        surrogate: Surrogate[ModelConfig],
        tracker: ModelTracker,
        metrics: Optional[List[str]] = None,
    ):
        """
        Args:
            surrogate: The surrogate model to evaluate.
            tracker: The collector from which to obtain the data for evaluation.
            metrics: The metrics to evaluate. If not provided, evaluates all metrics.
        """
        self.surrogate = surrogate
        self.tracker = tracker
        self.metrics = metrics

    def run(self) -> pd.DataFrame:
        """
        Runs the evaluation on the surrogate by applying LOOCV on the datasets
        being trained on. Metrics are then provided per test dataset.

        Returns:
            A data frame with the results for each fold, the metrics being the columns. The rows
                are indexed by the dataset which was left out.
        """
        if isinstance(self.surrogate, AutoGluonSurrogate):
            metrics = [
                self._run_on_dataset(x)
                for x in tqdm(list(loocv_split(self.tracker)))
            ]
        else:
            data = list(loocv_split(self.tracker))
            metrics = run_parallel(
                self._run_on_dataset,
                data=data,
                num_processes=min(
                    num_fitting_processes(
                        cpus_per_process=self.surrogate.required_cpus,
                        memory_per_process=self.surrogate.required_memory,
                    ),
                    len(data),
                ),
            )
        return pd.concat(metrics).set_index("test_dataset")

    def _run_on_dataset(
        self,
        data: Tuple[
            Tuple[List[Config[ModelConfig]], List[Performance]],
            Tuple[List[Config[ModelConfig]], List[Performance]],
        ],
    ) -> pd.DataFrame:
        (X_train, y_train), (X_test, y_test) = data

        # Fit model and predict
        self.surrogate.fit(X_train, y_train)
        y_pred = self.surrogate.predict(X_test)

        # Compute metrics
        scores = self._score(y_pred, y_test)
        return scores.assign(test_dataset=X_test[0].dataset.name())

    def _score(
        self, y_pred: List[Performance], y_true: List[Performance]
    ) -> pd.DataFrame:
        df_pred = Performance.to_dataframe(y_pred)
        df_true = Performance.to_dataframe(y_true)

        if self.metrics is not None:
            df_pred = df_pred[self.metrics]
            df_true = df_true[self.metrics]

        # We extract the NumPy arrays so that indexing is easier. Each metric is computed such that
        # it results in an array of shape [D] where D is the number of metrics.
        columns = df_pred.columns
        y_pred_min = df_pred.to_numpy()  # type: ignore
        y_true_min = df_true.to_numpy()  # type: ignore

        # Return all results
        metrics = {
            "nrmse": nrmse(y_pred_min, y_true_min),
            "smape": smape(y_pred_min, y_true_min),
            "mrr": mrr(y_pred_min, y_true_min),
            **{
                f"precision_{k}": precision_k(k, y_pred_min, y_true_min)
                for k in (5, 10, 20)
            },
            "ndcg": ndcg(y_pred_min, y_true_min),
        }
        column_index = pd.MultiIndex.from_tuples(
            [(c, m) for m in sorted(metrics) for c in columns]
        )
        values = np.concatenate([metrics[m] for m in sorted(metrics)])
        return pd.DataFrame(np.reshape(values, (1, -1)), columns=column_index)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Tuple, Optional
import math

import torch as pt
from torch import Tensor, BoolTensor
from torch import nn
from torch.nn import init
from torch.nn.functional import linear

from .base import Attention
from .posemb import SinusoidalPositionalEmbedding


class InterAttention(Attention):
    """
    Inter-attention module with k,v from source and q from target.

    Parameters
    ----------
    d_query : int
        query/decoder dimension
    d_kv: int
        key-value/encoder dimension
    n_head : int, optional
        number of attention heads, by default 1
    bias : bool, optional
        add bias term in input and output projections, by default True
    apply_kv_proj : bool, optional
        if True, apply linear projection to kv tensor and d_hidden = d_query,
            attention is performed in query feature space;
        if False, no linear porjection is applied and d_hidden = d_kv,
            attention is performed in key-value feature space;
            note that if n_head=1 in this case, it reduces to Luong's dot attention;
        by default True
    apply_rel_dist : Optional[str], optional
        add relative distance embeddings to dot-product attention, can be
            'add' (linearly combine key and dist),
            'dot' (dot product between key and dist),
            or None (disabled),
        by default None
    share_values : bool, optional
        if True, a value reprensentation is shared by all attention heads;
        only works when apply_kv_proj is False;
        by default False
        ref. https://arxiv.org/abs/1912.09363
    dropout : float, optional
        dropout rate, by default 0.0
    temperature : float, optional
        softmax temperature, by default 1.0
    """

    def __init__(
        self,
        d_query: int,
        d_kv: int,
        n_head: int = 1,
        bias: bool = True,
        apply_kv_proj: bool = True,
        apply_rel_dist: Optional[str] = None,
        share_values: bool = False,
        dropout: float = 0.0,
        temperature: float = 1.0,
    ):
        super(InterAttention, self).__init__(
            d_hidden=d_query if apply_kv_proj else d_kv,
            n_head=n_head,
            bias=bias,
            dropout=dropout,
            temperature=temperature,
        )
        self.apply_kv_proj = apply_kv_proj
        self.apply_rel_dist = apply_rel_dist
        self.share_values = share_values

        self._query_proj_weight = nn.Parameter(Tensor(self.d_hidden, d_query))
        if self.apply_kv_proj:
            self._kv_proj_weight = nn.Parameter(
                Tensor(2 * self.d_hidden, d_kv)
            )
            self._out_proj_weight = nn.Parameter(
                Tensor(self.d_hidden, self.d_hidden)
            )
        else:
            self.register_parameter("_kv_proj_weight", None)
            self.register_parameter("_out_proj_weight", None)
        if self.bias:
            self._query_proj_bias = nn.Parameter(Tensor(self.d_hidden))
            if self.apply_kv_proj:
                self._kv_proj_bias = nn.Parameter(Tensor(2 * self.d_hidden))
                self._out_proj_bias = nn.Parameter(Tensor(self.d_hidden))
            else:
                self.register_parameter("_kv_proj_bias", None)
                self.register_parameter("_out_proj_bias", None)
        else:
            self.register_parameter("_query_proj_bias", None)
            self.register_parameter("_kv_proj_bias", None)
            self.register_parameter("_out_proj_bias", None)
        if self.apply_rel_dist is None:
            self.posemb = None
            self.register_parameter("_pos_proj_weight", None)
        else:
            assert self.apply_rel_dist in ["dot", "add"]
            self.posemb = SinusoidalPositionalEmbedding(self.d_hidden)
            self._pos_proj_weight = nn.Parameter(
                Tensor(self.d_hidden, self.d_hidden)
            )
            if self.apply_rel_dist == "add":
                self._content_bias_weight = nn.Parameter(
                    Tensor(n_head, 1, self.d_head)
                )
                self._position_bias_weight = nn.Parameter(
                    Tensor(n_head, 1, self.d_head)
                )
            else:
                self.register_parameter("_content_bias_weight", None)
                self.register_parameter("_position_bias_weight", None)
        self._reset_parameters()

    def _reset_parameters(self):
        init.xavier_uniform_(self._query_proj_weight)
        if self.apply_kv_proj:
            if self.share_values:
                d_kv = self._kv_proj_weight.size(1)
                del self._kv_proj_weight
                self._key_proj_weight = nn.Parameter(
                    Tensor(self.d_hidden, d_kv)
                )
                self._value_proj_weight = nn.Parameter(
                    Tensor(self.d_head, d_kv)
                )
                init.xavier_uniform_(self._key_proj_weight)
                init.xavier_uniform_(self._value_proj_weight)
                self._kv_proj_weight = pt.cat(
                    [
                        self._key_proj_weight,
                        self._value_proj_weight.repeat(self.n_head, 1),
                    ],
                    dim=0,
                )
            else:
                init.xavier_uniform_(self._kv_proj_weight)
            init.xavier_uniform_(self._out_proj_weight)
        if self.bias:
            init.zeros_(self._query_proj_bias)
            if self.apply_kv_proj:
                init.zeros_(self._kv_proj_bias)
                init.zeros_(self._out_proj_bias)
        if self.apply_rel_dist is not None:
            init.xavier_uniform_(self._pos_proj_weight)
            if self.apply_rel_dist == "add":
                init.xavier_uniform_(self._content_bias_weight)
                init.xavier_uniform_(self._position_bias_weight)

    def _compute_qkv(
        self, query: Tensor, src: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor]:
        # [batch, qlen, d_hidden]
        q = linear(query, self._query_proj_weight, self._query_proj_bias)
        q = self._split_head(q)
        if self.apply_kv_proj:
            # [batch, klen, 2 * d_hidden]
            kv = linear(src, self._kv_proj_weight, self._kv_proj_bias)
            k, v = kv.chunk(2, dim=-1)
            k = self._split_head(k)
            v = self._split_head(v)
        else:
            k = v = self._split_head(src.clone())
        return q, k, v

    def _apply_mask(self, score: Tensor, key_mask: Optional[BoolTensor]):
        if key_mask is not None:
            key_mask = key_mask.unsqueeze(dim=1)  # head
            key_mask = key_mask.unsqueeze(dim=2)  # query
            score = score.masked_fill(key_mask, float("-inf"))
        return score

    def _compute_attn_score(
        self, q: Tensor, k: Tensor, key_mask: Optional[BoolTensor], gap: int
    ) -> Tensor:
        score = q.matmul(k.transpose(-1, -2))
        if self.apply_rel_dist:
            qlen = q.size(2)
            klen = k.size(2)
            # idx[i][j] = i-j+klen
            idx = (
                pt.arange(qlen).view(-1, 1)
                - pt.arange(klen).view(1, -1)
                + klen
                + (gap - 1)
            )
            idx = idx.to(device=q.device, dtype=pt.long)
            # torch.gather requires consistent shape
            # idx.shape = [batch, n_head, qlen, klen]
            idx = idx.view(1, 1, qlen, klen).expand(
                k.size(0), k.size(1), -1, -1
            )
            # r.shape = [1, qlen+klen, d_hidden]
            r = pt.arange(qlen + klen, dtype=pt.float, device=k.device).view(
                1, -1
            )
            r = r.add_(gap)
            r = linear(self.posemb(r), self._pos_proj_weight)
            # r.shape = [1, n_head, qlen+klen-1, d_head]
            r = self._split_head(r)
            if self.apply_rel_dist == "dot":
                # s_{ij} = <r_{i-j+klen}, (q_i+k_j)>
                #        = <q_i, r_{i-j+klen}> + <r_{i-j+klen}, k_j>

                # qr_{ij} = <q_i, r_j>
                # qr'_{ij} = qr_{i,idx[i][j]} = qr_{i,i-j+klen}
                qr = q.matmul(r.transpose(-1, -2))
                qr = qr.gather(dim=-1, index=idx)
                # rk_{ij} = <r_i, k_j>
                # rk'_{ij} = rk_{idx[i][j], j} = rk_{i-j+klen, j}
                rk = r.matmul(k.transpose(-1, -2))
                rk = rk.gather(dim=-2, index=idx)
                # s_{ij} = qr_{i,i-j+klen} + rk_{i-j+klen,j}
                s = qr + rk
            else:
                # transformer-xl style: https://arxiv.org/abs/1901.02860
                # s_{ij} = <q_i, r_{i-j+klen}> + <u, k_j> + <v, r_{i-j+klen}>
                #      u = _content_bias_weight
                #      v = _position_bias_weight

                # qr_{ij} = <q_i, r_j>
                # qr'_{ij} = qr_{i,idx[i][j]} = qr_{i,i-j+klen}
                qr = q.matmul(r.transpose(-1, -2))
                qr = qr.gather(dim=-1, index=idx)
                # rk_{ij} = <v, r_i> + <u, k_j>
                # rk'_{ij} = rk_{idx[i][j], j} = rk_{i-j+klen, j}
                rk = r.matmul(
                    self._content_bias_weight.transpose(-1, -2)
                ) + self._position_bias_weight.matmul(k.transpose(-1, -2))
                rk = rk.gather(dim=-2, index=idx)
                # s_{ij} = qr_{i,i-j+klen} + rk_{i-j+klen, j}
                s = qr + rk
            # add relative positional bias to content-based attention score
            score = score + s

        score = self._apply_mask(score, key_mask)
        score = score.div(math.sqrt(self.d_head)).div(self.temperature)
        score = score.softmax(dim=-1)
        score = self.dropout(score)
        return score

    def _compute_attn_output(self, score: Tensor, v: Tensor) -> Tensor:
        # values [batch, qlen, d_hidden]
        v = self._merge_head(score.matmul(v))
        if self.apply_kv_proj:
            v = linear(v, self._out_proj_weight, self._out_proj_bias)
        return v

    def forward(
        self,
        query: Tensor,
        src: Tensor,
        src_mask: Optional[BoolTensor] = None,
        gap: int = 1,
    ) -> Tensor:
        q, k, v = self._compute_qkv(query, src)
        score = self._compute_attn_score(q, k, key_mask=src_mask, gap=gap)
        v = self._compute_attn_output(score, v)
        return v



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Tuple, Optional
import math

import torch as pt
from torch import Tensor
from torch import nn


class Attention(nn.Module):
    """
    Base class of attention modules.

    *NOTE*: d_hidden must be divisible by n_head

    Parameters
    ----------
    d_hidden : int
        hidden dimension
    n_head : int, optional
        number of attention heads, by default 1
    bias : bool, optional
        add bias term in input and output projections, by default True
    dropout : float, optional
        dropout rate, by default 0.0
    temperature : float, optional
        softmax temperature, by default 1.0
    """

    def __init__(
        self,
        d_hidden: int,
        n_head: int = 1,
        bias: bool = True,
        dropout: float = 0.0,
        temperature: float = 1.0,
    ):
        super(Attention, self).__init__()
        if d_hidden % n_head != 0:
            raise ValueError(
                f"hidden dim {d_hidden} cannot be split into {n_head} heads"
            )
        self.d_hidden = d_hidden
        self.n_head = n_head
        self.d_head = d_hidden // n_head
        self.bias = bias
        self.dropout = nn.Dropout(dropout)
        self.temperature = temperature

    def _split_head(self, x: Tensor) -> Tensor:
        """
        Split hidden state into multi-heads.

        Args
        ----
            x : Tensor [batch, length, d_hidden]

        Returns
        -------
            Tensor [batch, n_head, length, d_head]
        """
        return (
            x.view(*x.shape[:-1], self.n_head, -1).transpose(1, 2).contiguous()
        )

    def _merge_head(self, x: Tensor) -> Tensor:
        """
        Merge multi-heads into one hidden state.

        Args
        ----------
            x : Tensor [batch, n_head, length, d_head]

        Returns
        -------
            Tensor [batch, length, d_hidden]
        """
        x = x.transpose(1, 2).contiguous()
        return x.view(*x.shape[:-2], -1)

    def _compute_qkv(self, *args, **kwargs):
        raise NotImplementedError

    def _compute_attn_score(self, q: Tensor, k: Tensor, **kwargs) -> Tensor:
        raise NotImplementedError

    def _compute_attn_output(
        self, score: Tensor, v: Tensor, **kwargs
    ) -> Tensor:
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import torch
import torch.nn.functional as F
from torch import nn

from torch.distributions.normal import Normal


class GaussianModel(nn.Module):
    r"""
    Model to learn a univariate Gaussian distribution.

    Arguments
    ----------
    mu: Mean of the Gaussian distribution
    sigma: Standard deviation of the Gaussian distribution
    device: The torch.device to use, typically cpu or gpu id
    """

    def __init__(self, mu, sigma, device=None):
        super().__init__()
        if device is not None:
            self.device = device
            mu = mu.to(device)
            sigma = sigma.to(device)
        self.mu = mu
        self.sigma = sigma
        self.distr = Normal(self.mu, self.sigma)

    def to_device(self, device):
        """
        Moves members to a specified torch.device.
        """
        self.device = device

    def forward(self, x):
        """
        Takes input x as new distribution parameters.
        """
        # If mini-batching
        if len(x.shape) > 1:
            self.mu_batch = x[:, 0]
            self.sigma_batch = F.softplus(x[:, 1])

        # If not mini-batching
        else:
            self.mu = x[0]
            self.distr = Normal(self.mu, self.sigma)

        return self.distr

    def log_prob(self, x):
        x = x.view(x.shape.numel())
        if x.shape[0] == 1:
            return self.distr.log_prob(x[0]).view(1)

        log_like_arr = torch.ones_like(x)
        for i in range(len(x)):
            self.mu = self.mu_batch[i]
            self.distr = Normal(self.mu, self.sigma)
            lpxx = self.distr.log_prob(x[i]).view(1)
            log_like_arr[i] = lpxx

        lpx = log_like_arr
        return lpx

    def icdf(self, value):
        return self.distr.icdf(value)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from traffic import *
from synthetic import *
from exchange_rate import *
from electricity import *

# from algo_clustering import *

# Please uncomment the following four commands to generate the corresponding dataset.
# The four lines corresponds to sythetic, traffic, exchange rate, electricity datasets, respectively.

# ===================================================================================================
get_mixed_pattern(unit_length=24, num_duplicates=2000)
group_traffic_cv(
    num_ts=800,
    num_groups=49,
    context_length=72,
    prediction_length=24,
    file_name="traffic",
)
group_exchangerate_cv(
    num_ts=8,
    num_groups=32,
    context_length=8,
    prediction_length=1,
    file_name="exchange_rate",
)
group_electricity_cv(
    num_ts=300,
    num_groups=70,
    context_length=72,
    prediction_length=24,
    file_name="electricity",
)
# ===================================================================================================


# (Ignore) the following are auxiliary functions for other datasets that are not included in the paper

# ===================================================================================================
# KMeans_inside_dataset(num_ts_=10, num_groups=32, file_name='traffic_toy32')
# KMeans_m5_dataset(num_ts_=500, num_groups=8, file_name='m5_10000')
# get_m4_by_freq(context_length=72, prediction_length=24, len_per_ts=200, num_ts=50)
# get_temperature_data(context_length=12, prediction_length=4, samples_per_ts=1000)
# ===================================================================================================



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict
import numpy as np
import pandas as pd
import scipy.stats as st


def compute_ranks(candidates: Dict[str, pd.DataFrame]) -> pd.DataFrame:
    """
    Computes the ranks of the provided candidates for each column in the
    provided dataframe. If candidates show the same values, ranks are tied.

    Args:
        candidates: A mapping from candidate names to performances. The indexes of all candidates
            must align and ranks are computed for each index value.

    Returns:
        A data frame containing the results of all candidates and datasets. The data frame has a
            multi-index with the candidate name on the first level and the dataset on the second.
    """
    _check_candidates(candidates)
    ref = next(iter(candidates.values()))

    # Sort data frames such that we can use NumPy operations to compute ranks
    sorted_candidates = [
        {"name": k, "df": d[ref.columns].sort_index()}
        for k, d in candidates.items()
    ]
    arr = np.stack([c["df"].to_numpy() for c in sorted_candidates], axis=0)  # type: ignore
    ranks = st.rankdata(arr, axis=0, method="min")

    # Extract data frames from ranks again
    sorted_index = ref.index.sort_values()

    index = pd.MultiIndex.from_tuples(
        [
            t
            for candidate in sorted_candidates
            for t in [(candidate["name"], item) for item in sorted_index]
        ],
        names=["candidate", "example"],
    )
    return pd.DataFrame(
        np.concatenate(ranks), index=index, columns=ref.columns
    )


def _check_candidates(candidates: Dict[str, pd.DataFrame]) -> None:
    ref = next(iter(candidates.values()))
    assert all(
        set(c.index) == set(ref.index) for c in candidates.values()
    ), "Candidates have differing indices."
    assert all(
        set(c.columns) == set(ref.columns) for c in candidates.values()
    ), "Candidates have differing columns."



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, Optional, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from torch.distributions import (
    Beta,
    Distribution,
    Gamma,
    Laplace,
    Normal,
    Poisson,
)

from src.gluonts.core.component import validated
from src.gluonts.model.forecast_generator import (
    DistributionForecastGenerator,
    ForecastGenerator,
)
from src.gluonts.torch.distributions import AffineTransformed

from .output import Output


class DistributionOutput(Output):
    r"""
    Class to construct a distribution given the output of a network.
    """

    distr_cls: type

    @validated()
    def __init__(self, beta: float = 0.0) -> None:
        self.beta = beta

    def _base_distribution(self, distr_args):
        return self.distr_cls(*distr_args)

    def distribution(
        self,
        distr_args,
        loc: Optional[torch.Tensor] = None,
        scale: Optional[torch.Tensor] = None,
    ) -> Distribution:
        r"""
        Construct the associated distribution, given the collection of
        constructor arguments and, optionally, a scale tensor.

        Parameters
        ----------
        distr_args
            Constructor arguments for the underlying Distribution type.
        loc
            Optional tensor, of the same shape as the
            batch_shape+event_shape of the resulting distribution.
        scale
            Optional tensor, of the same shape as the
            batch_shape+event_shape of the resulting distribution.
        """
        distr = self._base_distribution(distr_args)
        if loc is None and scale is None:
            return distr
        else:
            return AffineTransformed(distr, loc=loc, scale=scale)

    def loss(
        self,
        target: torch.Tensor,
        distr_args: Tuple[torch.Tensor, ...],
        loc: Optional[torch.Tensor] = None,
        scale: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        distribution = self.distribution(distr_args, loc=loc, scale=scale)
        nll = -distribution.log_prob(target)
        if self.beta > 0.0:
            variance = distribution.variance
            nll = nll * (variance.detach() ** self.beta)
        return nll

    @property
    def event_dim(self) -> int:
        r"""
        Number of event dimensions, i.e., length of the `event_shape` tuple, of
        the distributions that this object constructs.
        """
        return len(self.event_shape)

    def domain_map(self, *args: torch.Tensor):
        r"""
        Converts arguments to the right shape and domain.

        The domain depends on the type of distribution, while the correct shape
        is obtained by reshaping the trailing axis in such a way that the
        returned tensors define a distribution of the right event_shape.
        """
        raise NotImplementedError()

    @property
    def forecast_generator(self) -> ForecastGenerator:
        return DistributionForecastGenerator(self)


class NormalOutput(DistributionOutput):
    args_dim: Dict[str, int] = {"loc": 1, "scale": 1}
    distr_cls: type = Normal

    @classmethod
    def domain_map(cls, loc: torch.Tensor, scale: torch.Tensor):  # type: ignore
        scale = F.softplus(scale)
        return loc.squeeze(-1), scale.squeeze(-1)

    @property
    def event_shape(self) -> Tuple:
        return ()


class LaplaceOutput(DistributionOutput):
    args_dim: Dict[str, int] = {"loc": 1, "scale": 1}
    distr_cls: type = Laplace

    @classmethod
    def domain_map(cls, loc: torch.Tensor, scale: torch.Tensor):  # type: ignore
        scale = F.softplus(scale)
        return loc.squeeze(-1), scale.squeeze(-1)

    @property
    def event_shape(self) -> Tuple:
        return ()


class BetaOutput(DistributionOutput):
    args_dim: Dict[str, int] = {"concentration1": 1, "concentration0": 1}
    distr_cls: type = Beta

    @classmethod
    def domain_map(  # type: ignore
        cls, concentration1: torch.Tensor, concentration0: torch.Tensor
    ):
        epsilon = np.finfo(cls._dtype).eps  # machine epsilon
        concentration1 = F.softplus(concentration1) + epsilon
        concentration0 = F.softplus(concentration0) + epsilon
        return concentration1.squeeze(dim=-1), concentration0.squeeze(dim=-1)

    @property
    def event_shape(self) -> Tuple:
        return ()

    @property
    def value_in_support(self) -> float:
        return 0.5


class GammaOutput(DistributionOutput):
    args_dim: Dict[str, int] = {"concentration": 1, "rate": 1}
    distr_cls: type = Gamma

    @classmethod
    def domain_map(cls, concentration: torch.Tensor, rate: torch.Tensor):  # type: ignore
        epsilon = np.finfo(cls._dtype).eps  # machine epsilon
        concentration = F.softplus(concentration) + epsilon
        rate = F.softplus(rate) + epsilon
        return concentration.squeeze(dim=-1), rate.squeeze(dim=-1)

    @property
    def event_shape(self) -> Tuple:
        return ()

    @property
    def value_in_support(self) -> float:
        return 0.5


class PoissonOutput(DistributionOutput):
    args_dim: Dict[str, int] = {"rate": 1}
    distr_cls: type = Poisson

    @classmethod
    def domain_map(cls, rate: torch.Tensor):  # type: ignore
        rate_pos = F.softplus(rate).clone()
        return (rate_pos.squeeze(-1),)

    # Overwrites the parent class method. We cannot scale using the affine
    # transformation since Poisson should return integers. Instead we scale
    # the parameters.
    def distribution(
        self,
        distr_args,
        loc: Optional[torch.Tensor] = None,
        scale: Optional[torch.Tensor] = None,
    ) -> Distribution:
        (rate,) = distr_args

        if scale is not None:
            rate *= scale

        return Poisson(rate=rate)

    @property
    def event_shape(self) -> Tuple:
        return ()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from functools import partial
from typing import Callable, Iterator, List, Optional
from pathlib import Path

import mxnet as mx
import numpy as np


from src.gluonts.core.component import Type, validated
from src.gluonts.core.serde import load_json
from src.gluonts.dataset.common import DataEntry, Dataset
from src.gluonts.dataset.loader import DataBatch, InferenceDataLoader
from src.gluonts.model.forecast import Forecast
from src.gluonts.model.forecast_generator import (
    ForecastGenerator,
    SampleForecastGenerator,
)
from src.gluonts.mx.batchify import stack
from src.gluonts.mx.context import get_mxnet_context
from src.gluonts.mx.distribution import DistributionOutput, StudentTOutput
from src.gluonts.mx.model.deepar import DeepAREstimator
from src.gluonts.mx.model.deepar._network import DeepARPredictionNetwork
from src.gluonts.mx.model.predictor import RepresentableBlockPredictor
from src.gluonts.mx.trainer import Trainer
from src.gluonts.mx.util import import_repr_block
from src.gluonts.time_feature import TimeFeature
from src.gluonts.transform import InstanceSampler, Transformation
from src.gluonts.transform.feature import MissingValueImputation


LOSS_FUNCTIONS = ["crps_univariate", "nll"]


def batchify_with_dict(
    data: List[dict],
    ctx: Optional[mx.context.Context] = None,
    dtype: Optional[Type] = np.float32,
    variable_length: bool = False,
    is_right_pad: bool = True,
) -> DataBatch:
    return {
        key: (
            stack(
                data=[item[key] for item in data],
                ctx=ctx,
                dtype=dtype,
                variable_length=variable_length,
                is_right_pad=is_right_pad,
            )
            if not isinstance(data[0][key], dict)
            else batchify_with_dict(data=[item[key] for item in data])
        )
        for key in data[0].keys()
    }


class RepresentableBlockPredictorBatchifyWithDict(RepresentableBlockPredictor):
    """
    We need the stack function `batchify_with_dict` in order to pass the
    features at the aggregated level properly during prediction.

    Gluonts does not allow this without changing the line corresponding to the
    `InferenceDataLoader` in the `predict` function.
    """

    def __init__(
        self,
        prediction_net: mx.gluon.HybridBlock,
        batch_size: int,
        prediction_length: int,
        ctx: mx.Context,
        input_transform: Transformation,
        lead_time: int = 0,
        forecast_generator: ForecastGenerator = SampleForecastGenerator(),
        output_transform: Optional[
            Callable[[DataEntry, np.ndarray], np.ndarray]
        ] = None,
        dtype: Type = np.float32,
    ) -> None:
        super().__init__(
            prediction_net=prediction_net,
            batch_size=batch_size,
            prediction_length=prediction_length,
            ctx=ctx,
            input_transform=input_transform,
            lead_time=lead_time,
            forecast_generator=forecast_generator,
            output_transform=output_transform,
            dtype=dtype,
        )

    def predict(
        self,
        dataset: Dataset,
        num_samples: Optional[int] = None,
        num_workers: Optional[int] = None,
        num_prefetch: Optional[int] = None,
        **kwargs,
    ) -> Iterator[Forecast]:
        inference_data_loader = InferenceDataLoader(
            dataset,
            transform=self.input_transform,
            batch_size=self.batch_size,
            stack_fn=partial(
                batchify_with_dict, ctx=self.ctx, dtype=self.dtype
            ),
        )
        with mx.Context(self.ctx):
            yield from self.forecast_generator(
                inference_data_loader=inference_data_loader,
                prediction_net=self.prediction_net,
                input_names=self.input_names,
                output_transform=self.output_transform,
                num_samples=num_samples,
            )

    @classmethod
    def deserialize(
        cls, path: Path, ctx: Optional[mx.Context] = None
    ) -> "RepresentableBlockPredictorBatchifyWithDict":
        ctx = ctx if ctx is not None else get_mxnet_context()

        with mx.Context(ctx):
            # deserialize constructor parameters
            with (path / "parameters.json").open("r") as fp:
                parameters = load_json(fp.read())

            # deserialize transformation chain
            with (path / "input_transform.json").open("r") as fp:
                transform = load_json(fp.read())

            # deserialize prediction network
            prediction_net = import_repr_block(path, "prediction_net")

            # input_names is derived from the prediction_net
            if "input_names" in parameters:
                del parameters["input_names"]

            parameters["ctx"] = ctx
            return RepresentableBlockPredictorBatchifyWithDict(
                input_transform=transform,
                prediction_net=prediction_net,
                batch_size=parameters["batch_size"],
                prediction_length=parameters["prediction_length"],
                ctx=parameters["ctx"],
                dtype=parameters["dtype"],
            )


# Gluonts estimator should expose this function. Currently it has api for creating predictor but not prediction network!
def create_prediction_network(
    estimator: DeepAREstimator,
) -> DeepARPredictionNetwork:
    return DeepARPredictionNetwork(
        num_parallel_samples=estimator.num_parallel_samples,
        num_layers=estimator.num_layers,
        num_cells=estimator.num_cells,
        cell_type=estimator.cell_type,
        history_length=estimator.history_length,
        context_length=estimator.context_length,
        prediction_length=estimator.prediction_length,
        distr_output=estimator.distr_output,
        dropoutcell_type=estimator.dropoutcell_type,
        dropout_rate=estimator.dropout_rate,
        cardinality=estimator.cardinality,
        embedding_dimension=estimator.embedding_dimension,
        lags_seq=estimator.lags_seq,
        scaling=estimator.scaling,
        dtype=estimator.dtype,
        num_imputation_samples=estimator.num_imputation_samples,
        default_scale=estimator.default_scale,
        minimum_scale=estimator.minimum_scale,
        impute_missing_values=estimator.impute_missing_values,
    )


# Hack to expose history_length to constructor to allow for proper
# serialization
class DeepAREstimatorForCOP(DeepAREstimator):
    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        trainer: Trainer = Trainer(),
        context_length: Optional[int] = None,
        num_layers: int = 2,
        num_cells: int = 40,
        cell_type: str = "lstm",
        dropoutcell_type: str = "ZoneoutCell",
        dropout_rate: float = 0.1,
        use_feat_dynamic_real: bool = False,
        use_feat_static_cat: bool = False,
        use_feat_static_real: bool = False,
        cardinality: Optional[List[int]] = None,
        embedding_dimension: Optional[List[int]] = None,
        distr_output: DistributionOutput = StudentTOutput(),
        scaling: bool = True,
        lags_seq: Optional[List[int]] = None,
        time_features: Optional[List[TimeFeature]] = None,
        num_parallel_samples: int = 100,
        imputation_method: Optional[MissingValueImputation] = None,
        train_sampler: Optional[InstanceSampler] = None,
        validation_sampler: Optional[InstanceSampler] = None,
        dtype: Type = np.float32,
        alpha: float = 0.0,
        beta: float = 0.0,
        batch_size: int = 32,
        default_scale: Optional[float] = None,
        minimum_scale: float = 1e-10,
        impute_missing_values: bool = False,
        num_imputation_samples: int = 1,
        history_length: Optional[int] = None,
    ) -> None:
        super().__init__(
            freq=freq,
            prediction_length=prediction_length,
            trainer=trainer,
            context_length=context_length,
            num_layers=num_layers,
            num_cells=num_cells,
            cell_type=cell_type,
            dropoutcell_type=dropoutcell_type,
            dropout_rate=dropout_rate,
            use_feat_dynamic_real=use_feat_dynamic_real,
            use_feat_static_cat=use_feat_static_cat,
            use_feat_static_real=use_feat_static_real,
            cardinality=cardinality,
            embedding_dimension=embedding_dimension,
            distr_output=distr_output,
            scaling=scaling,
            lags_seq=lags_seq,
            time_features=time_features,
            num_parallel_samples=num_parallel_samples,
            imputation_method=imputation_method,
            train_sampler=train_sampler,
            validation_sampler=validation_sampler,
            alpha=alpha,
            beta=beta,
            default_scale=default_scale,
            minimum_scale=minimum_scale,
            impute_missing_values=impute_missing_values,
            num_imputation_samples=num_imputation_samples,
            batch_size=batch_size,
            dtype=dtype,
        )
        self.freq = freq
        self.history_length: int = (
            history_length
            if history_length is not None
            else self.history_length
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
from dataclasses import dataclass, asdict
from typing import Optional, Literal, Tuple, List, Dict
from pathlib import Path
import json
from tqdm import tqdm
import hashlib
import numpy as np
import torch
import torch.nn as nn
import pytorch_lightning as pl
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF
from torch.utils.data import DataLoader, ConcatDataset
from src.gluonts.time_feature import get_seasonality

from .artificial import _write_data_to_file
from meta.datasets.registry import register_data_module
from meta.data.batch import TripletBatch
from meta.datasets.splits import DatasetSplits
from meta.data.sampling import TripletDataset
from meta.common.torch import tensor_to_np
from meta.metrics.numpy import compute_metrics


@dataclass(frozen=True)
class CheatMetaData:
    """
    Meta data for a cheating dataset.
    """

    context_length_multiple: int
    support_length_multiple: int
    prediction_length: int
    test_set_size: int
    val_set_size: int
    train_set_size: int
    freq: str
    # below are the parameters that we will play in the artificial experiments
    support_set_size: int
    cheat_chance: float
    noise_level: float
    query_length_scale: float
    support_length_scale_same_as_query: bool

    @classmethod
    def parse_file(cls, file: Path):
        with open(file) as json_file:
            return CheatMetaData(**json.load(json_file))

    def save(self, file: Path):
        with open(file, "w") as fp:
            json.dump(asdict(self), fp)

    def get_hash(self):
        return (
            hashlib.md5(json.dumps(asdict(self)).encode("utf-8"))
            .digest()
            .hex()
        )


@register_data_module
class CheatArtificialDataModule(pl.LightningDataModule):
    """
    A data module which provides datasets with different ground truth / counter
    factual injection modes for the support set.

    Parameters
    ----------
    context_length_multiple: int
        Context length is prediction length x ``context_length_multiple``.
    prediction_length: int
        Number of predicted time steps.
    support_length_multiple: int
        Support time series length is prediction length x ``support_length_multiple``.
    test_set_size: int
        The size of the test split of the generated dataset.
    val_set_size: int
        The size of the val split of the generated dataset.
    train_set_size: int
        The size of the train split of the generated dataset.
    freq: str
        Not clear if this is actually used here on not.
    support_set_size: int
        Number of time series in the support set.
    data_dir: pathlib.Path
        Path to data.
    cheat_chance: float
        The chance of injecting ground truth in a support set.
    noise_level: float
        The level of noise added to the ground truth when cheating.
    query_length_scale: float
        The length scale used by the gaussian process to sample the queries from.
    support_length_scale_same_as_query: bool
        Whether the length scale of the gaussian process for query and support set match.
    standardize: bool
        If each time series is standardized.
    batch_size_train: int
        Batch size during training.
    batch_size_val_test: int
        Batch size during validation and testing.
    num_workers: int
        Number of workers.
    random_seed: int
        The random seed.
    """

    def __init__(
        self,
        # dataset metadata config starts
        context_length_multiple: int = 4,
        prediction_length: int = 24,
        support_length_multiple: int = 4,
        test_set_size: int = 500,
        val_set_size: int = 500,
        train_set_size: int = 10000,
        freq: str = "H",
        support_set_size: int = 3,
        cheat_chance: float = 1.0,
        noise_level: float = 0.0,
        query_length_scale: float = 5.0,
        support_length_scale_same_as_query: bool = True,
        # data loader config starts
        data_dir: Path = Path.home() / ".mxnet" / "gluon-ts",
        standardize: bool = False,
        batch_size_train: int = 128,
        batch_size_val_test: int = 1000,
        num_workers: int = 0,
        random_seed: int = int.from_bytes(os.urandom(3), byteorder="big"),
    ):
        assert 0.1 <= query_length_scale <= 10

        super().__init__()
        # parameters of the dataset
        self.meta = CheatMetaData(
            context_length_multiple=context_length_multiple,
            prediction_length=prediction_length,
            support_length_multiple=support_length_multiple,
            test_set_size=test_set_size,
            val_set_size=val_set_size,
            train_set_size=train_set_size,
            freq=freq,
            support_set_size=support_set_size,
            cheat_chance=cheat_chance,
            noise_level=noise_level,
            query_length_scale=query_length_scale,
            support_length_scale_same_as_query=support_length_scale_same_as_query,
        )
        # parameters of the loader
        self.data_dir = data_dir
        self.standardize = standardize
        self.batch_size_train = batch_size_train
        self.batch_size_val_test = batch_size_val_test
        self.num_workers = num_workers
        self.splits: DatasetSplits
        self.random_state = np.random.RandomState(random_seed)

    @property
    def dataset_names_val_test(self) -> List[str]:
        return list({"cc1.0", f"cc{self.meta.cheat_chance}"})

    @property
    def dataset_names_val(self) -> List[str]:
        return self.dataset_names_val_test

    @property
    def dataset_names_test(self) -> List[str]:
        return self.dataset_names_val_test

    @property
    def context_length(self) -> int:
        return self.meta.context_length_multiple * self.meta.prediction_length

    @property
    def support_length(self) -> int:
        return self.meta.support_length_multiple * self.meta.prediction_length

    @property
    def prediction_length(self) -> int:
        return self.meta.prediction_length

    @property
    def dataset_name(self) -> str:
        return "cheat_" + self.meta.get_hash()

    @property
    def root(self) -> Path:
        """
        Returns the directory where all the data pertaining to this dataset is
        stored.
        """
        return self.data_dir / "artificial" / self.dataset_name

    def setup(self, stage: Optional[str] = None) -> None:
        self.generate()
        self.splits = DatasetSplits(
            self.meta,
            self.root,
            self.dataset_name,
            self.prediction_length,
            self.standardize,
        )
        self.splits.train().prepare()
        for d_name in self.dataset_names_val:
            self.splits.val(d_name).prepare()
        for d_name in self.dataset_names_test:
            self.splits.test(d_name).prepare()

    def train_dataloader(self) -> DataLoader[TripletBatch]:
        split = self.splits.train()
        return DataLoader(
            TripletDataset(
                queries=split.data(), support_sets=split.support_set()
            ),
            collate_fn=TripletBatch.collate,
            batch_size=self.batch_size_train,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def val_dataloader(self) -> DataLoader[TripletBatch]:
        splits = [self.splits.val(d_name) for d_name in self.dataset_names_val]

        return list(
            [
                DataLoader(
                    TripletDataset(
                        queries=split.data(), support_sets=split.support_set()
                    ),
                    collate_fn=TripletBatch.collate,
                    batch_size=self.batch_size_val_test,
                    num_workers=self.num_workers,
                    pin_memory=True,
                )
                for split in splits
            ]
        )

    def test_dataloader(self) -> DataLoader[TripletBatch]:
        splits = [
            self.splits.test(d_name) for d_name in self.dataset_names_test
        ]
        return list(
            [
                DataLoader(
                    TripletDataset(
                        queries=split.data(), support_sets=split.support_set()
                    ),
                    collate_fn=TripletBatch.collate,
                    batch_size=self.batch_size_val_test,
                    num_workers=self.num_workers,
                    pin_memory=True,
                )
                for split in splits
            ]
        )

    def generate(self) -> None:
        if self.root.exists():
            return

        self.root.mkdir(parents=True, exist_ok=True)
        self.meta.save(self.root / "metadata.json")
        for split, n_samples in zip(
            ["train", "val", "test"],
            [
                self.meta.train_set_size,
                self.meta.val_set_size,
                self.meta.test_set_size,
            ],
        ):
            if split in ["val", "test"]:
                self.generate_split(
                    split + "_cc1.0", n_samples, always_cheat=True
                )
                self.generate_split(
                    split + f"_cc{self.meta.cheat_chance}",
                    n_samples,
                    always_cheat=False,
                )
            else:
                self.generate_split(split, n_samples, always_cheat=False)

    def generate_split(
        self,
        split: Literal["train", "val", "test"],
        n_samples: int,
        always_cheat=True,
        query_length_scale=5.0,
    ) -> None:
        queries, support_sets = zip(
            *[
                self.generate_artificial_tuplets(
                    item_id=i,
                    always_cheat=always_cheat,
                    query_length_scale=query_length_scale,
                )
                for i in tqdm(
                    range(n_samples), desc="generating artificial data"
                )
            ]
        )
        _write_data_to_file(self.root / split / "data.json", queries)
        _write_data_to_file(
            self.root / split / ".support_set.json", support_sets
        )

    def generate_artificial_tuplets(
        self, item_id: int, always_cheat: bool, query_length_scale: float
    ) -> Tuple:
        # generate query
        query_length = self.meta.prediction_length + self.context_length
        query_x = np.arange(query_length).reshape(-1, 1)
        support_x = np.arange(self.support_length).reshape(-1, 1)

        kernel = 0.5 * RBF(
            length_scale=query_length_scale, length_scale_bounds=(1e-1, 10.0)
        )
        gpr = GaussianProcessRegressor(
            kernel=kernel, random_state=self.random_state
        )
        query = gpr.sample_y(
            query_x, 1, random_state=self.random_state
        ).squeeze()
        assert query.shape == (query_length,), query.shape

        # prepare support set
        if self.meta.support_length_scale_same_as_query:
            support_set = gpr.sample_y(
                support_x,
                self.meta.support_set_size,
                random_state=self.random_state,
            ).T
        else:
            ys = []
            for _ in range(self.meta.support_set_size):
                sample_length_scale = self.random_state.uniform(0, 10)
                kernel = 0.5 * RBF(
                    length_scale=sample_length_scale,
                    length_scale_bounds=(1e-1, 10.0),
                )
                gpr_support = GaussianProcessRegressor(
                    kernel=kernel, random_state=self.random_state
                )
                y_hat = gpr_support.sample_y(support_x)
                ys.append(y_hat.ravel())
            support_set = np.array(ys)

        assert support_set.shape == (
            self.meta.support_set_size,
            self.support_length,
        ), support_set.shape

        # decide to cheat on 1 TS in support set or not based on cheat_chance
        is_cheat = (
            True
            if always_cheat
            else self.random_state.rand() <= self.meta.cheat_chance
        )

        if is_cheat:
            # generate a random place to for TS in support set to start cheating
            marker_length = self.meta.prediction_length
            signal_length = marker_length + self.meta.prediction_length
            marker_start = self.random_state.choice(
                np.arange(0, self.support_length - signal_length)
            )

            # prepare data to train the GP
            x = np.array(
                range(marker_start, marker_start + signal_length)
            ).reshape(-1, 1)
            y = query.tolist()[-signal_length:]

            # fit gp and sample from fitted gp
            gpr.fit(x, y)

            # generate a random index from support set
            si = self.random_state.choice(self.meta.support_set_size)

            if self.meta.support_length_scale_same_as_query:
                y_hat = gpr.sample_y(support_x)
            else:
                sample_length_scale = self.random_state.uniform(0, 10)
                gpr.kernel_.set_params(
                    **{"k2__length_scale": sample_length_scale}
                )
                y_hat = gpr.sample_y(support_x)
            support_set[si] = y_hat.ravel()

        # add noise
        if self.meta.noise_level > 0:
            noise = (
                self.random_state.randn(query.shape[0]) * self.meta.noise_level
            )
            query += noise

            new_support_set = []
            for s in support_set:
                noise = (
                    self.random_state.randn(s.shape[0]) * self.meta.noise_level
                )
                new_support_set.append(s + noise)
            support_set = np.array(new_support_set)

        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s

    def get_log_batches(self, n_logging_samples: int) -> Tuple[TripletBatch]:
        log_batch_train = (
            next(iter(self.train_dataloader()))
            .reduce_to_unique_query()
            .first_n(n_logging_samples)
        )
        log_batch_val = (
            next(iter(self.val_dataloader()[0]))
            .reduce_to_unique_query()
            .first_n(n_logging_samples)
        )
        return log_batch_train, log_batch_val

    def evaluate_model(
        self,
        model: nn.Module,
        quantiles: List[str],
        test: bool = False,
    ) -> List[Dict[str, float]]:
        metrics = []

        self.setup()
        pred = []
        if test:
            split = self.splits.test()
            dl = self.test_dataloader()
        else:
            split = self.splits.val()
            dl = self.val_dataloader()

        for batch in tqdm(
            dl,
            desc=f"generating predictions for dataset {self.name()} with support set size",
        ):
            pred.append(model(supps=batch.support_set, query=batch.query_past))

        # redo standardization for evaluation
        pred = split.data().rescale_dataset(torch.cat(pred, dim=0))
        pred = tensor_to_np(pred)

        # use only the length that should be included for evaluation
        pred = pred[:, : self.prediction_length, ...]

        m = compute_metrics(
            pred,
            split.evaluation(),
            quantiles=quantiles,
            seasonality=get_seasonality(self.meta.freq),
        )
        metrics.append(m)
        return metrics

    @classmethod
    def name(cls) -> str:
        return "dm_cheat"


@dataclass(frozen=True)
class CheatCounterfactual(CheatMetaData):
    """
    Meta data for a cheating dataset.
    """

    counterfactual_size: int
    counterfactual_mixing: bool

    @classmethod
    def parse_file(cls, file: Path):
        with open(file) as json_file:
            return CheatCounterfactual(**json.load(json_file))


@register_data_module
class CheatCounterfactualArtificialDataModule(CheatArtificialDataModule):
    def __init__(
        self,
        # dataset metadata config starts
        context_length_multiple: int = 4,
        prediction_length: int = 24,
        support_length_multiple: int = 4,
        test_set_size: int = 500,
        val_set_size: int = 500,
        train_set_size: int = 10000,
        freq: str = "H",
        support_set_size: int = 3,
        cheat_chance: float = 1.0,
        noise_level: float = 0.0,
        query_length_scale: float = 5.0,
        support_length_scale_same_as_query: bool = True,
        counterfactual_size: int = 1,
        counterfactual_mixing: bool = False,
        # data loader config starts
        data_dir: Path = Path.home() / ".mxnet" / "gluon-ts",
        standardize: bool = False,
        batch_size_train: int = 128,
        batch_size_val_test: int = 1000,
        num_workers: int = 0,
        random_seed: int = int.from_bytes(os.urandom(3), byteorder="big"),
    ):
        assert 0.1 <= query_length_scale <= 10
        assert counterfactual_size < support_set_size
        super().__init__()

        # parameters of the dataset
        self.meta = CheatCounterfactual(
            context_length_multiple=context_length_multiple,
            prediction_length=prediction_length,
            support_length_multiple=support_length_multiple,
            test_set_size=test_set_size,
            val_set_size=val_set_size,
            train_set_size=train_set_size,
            freq=freq,
            support_set_size=support_set_size,
            cheat_chance=cheat_chance,
            noise_level=noise_level,
            query_length_scale=query_length_scale,
            support_length_scale_same_as_query=support_length_scale_same_as_query,
            counterfactual_size=counterfactual_size,
            counterfactual_mixing=counterfactual_mixing,
        )

        # parameters of the loader
        self.data_dir = data_dir
        self.standardize = standardize
        self.batch_size_train = batch_size_train
        self.batch_size_val_test = batch_size_val_test
        self.num_workers = num_workers
        self.splits: DatasetSplits
        self.random_state = np.random.RandomState(random_seed)

    @property
    def dataset_names_val_test(self) -> List[str]:
        return list({"cc1.0", f"cf{self.meta.counterfactual_size}"})

    @property
    def dataset_names_train(self) -> List[str]:
        if self.meta.counterfactual_size == 0:
            return ["cc1.0"]
        else:
            if self.meta.counterfactual_mixing:
                return ["cc1.0", f"cf{self.meta.counterfactual_size}"]
            else:
                return [f"cf{self.meta.counterfactual_size}"]

    @property
    def dataset_name(self) -> str:
        return "cf_" + self.meta.get_hash()

    def train_dataloader(self) -> DataLoader[TripletBatch]:
        splits = [
            self.splits.train(name=d_name)
            for d_name in self.dataset_names_train
        ]

        datasets = ConcatDataset(
            [
                TripletDataset(
                    queries=split.data(), support_sets=split.support_set()
                )
                for split in splits
            ]
        )

        return DataLoader(
            datasets,
            shuffle=True,
            collate_fn=TripletBatch.collate,
            batch_size=self.batch_size_train,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def generate(self) -> None:
        if self.root.exists():
            return

        self.root.mkdir(parents=True, exist_ok=True)
        self.meta.save(self.root / "metadata.json")
        for split, n_samples in zip(
            ["train", "val", "test"],
            [
                self.meta.train_set_size,
                self.meta.val_set_size,
                self.meta.test_set_size,
            ],
        ):
            self.generate_split(
                split + "_cc1.0",
                n_samples,
                always_cheat=True,
                counterfactual_size=0,
            )
            self.generate_split(
                f"{split}_cf{self.meta.counterfactual_size}",
                n_samples,
                always_cheat=True,
                counterfactual_size=self.meta.counterfactual_size,
            )

    def generate_split(
        self,
        split: str,
        n_samples: int,
        always_cheat=True,
        query_length_scale: float = 5.0,
        counterfactual_size: int = 0,
    ) -> None:
        queries, support_sets = zip(
            *[
                self.generate_artificial_tuplets(
                    item_id=i,
                    always_cheat=always_cheat,
                    query_length_scale=query_length_scale,
                    counterfactual_size=counterfactual_size,
                )
                for i in tqdm(
                    range(n_samples), desc="generating artificial data"
                )
            ]
        )
        _write_data_to_file(self.root / split / "data.json", queries)
        _write_data_to_file(
            self.root / split / ".support_set.json", support_sets
        )

    def generate_artificial_tuplets(
        self,
        item_id: int,
        always_cheat: bool,
        query_length_scale: float,
        counterfactual_size: int,
    ) -> Tuple:
        assert always_cheat is True, "In CounterFactual, we always cheat"
        assert self.meta.support_length_scale_same_as_query is True

        # generate query
        query_length = self.meta.prediction_length + self.context_length
        query_x = np.arange(query_length).reshape(-1, 1)
        support_x = np.arange(self.support_length).reshape(-1, 1)

        kernel = 0.5 * RBF(
            length_scale=query_length_scale, length_scale_bounds=(1e-1, 10.0)
        )
        gpr = GaussianProcessRegressor(
            kernel=kernel, random_state=self.random_state
        )
        query = gpr.sample_y(
            query_x, 1, random_state=self.random_state
        ).squeeze()
        assert query.shape == (query_length,), query.shape

        support_set = gpr.sample_y(
            support_x,
            self.meta.support_set_size,
            random_state=self.random_state,
        ).T

        assert support_set.shape == (
            self.meta.support_set_size,
            self.support_length,
        ), support_set.shape

        # generate a random place to for TS in support set to start cheating
        marker_length = self.meta.prediction_length
        signal_length = marker_length + self.meta.prediction_length
        marker_start = self.random_state.choice(
            np.arange(0, self.support_length - signal_length)
        )

        # prepare data to train the GP
        x = np.array(
            range(marker_start, marker_start + signal_length)
        ).reshape(-1, 1)
        y = query.tolist()[-signal_length:]

        # fit gp and sample from fitted gp
        gpr.fit(x, y)

        # generate a random index from support set to cheat
        si = self.random_state.choice(self.meta.support_set_size)
        y_hat = gpr.sample_y(support_x, 1, random_state=self.random_state)
        support_set[si] = y_hat.ravel()

        indices = list(range(self.meta.support_set_size))
        indices.remove(si)

        # generate a counterfactual TS, only contains information in the context (marker length)
        x_cf = np.array(
            range(marker_start, marker_start + marker_length)
        ).reshape(-1, 1)
        y_cf = query.tolist()[-signal_length:-marker_length]

        # fit gp and sample from fitted gp
        for _ in range(counterfactual_size):
            gpr_cf = GaussianProcessRegressor(
                kernel=kernel, random_state=self.random_state
            )
            gpr_cf.fit(x_cf, y_cf)
            # generate a random index from support set for cf
            cf_si = self.random_state.choice(indices)
            y_hat_cf = gpr_cf.sample_y(
                support_x, 1, random_state=self.random_state
            )
            support_set[cf_si] = y_hat_cf.ravel()
            indices.remove(cf_si)

        # add noise
        if self.meta.noise_level > 0:
            noise = (
                self.random_state.randn(query.shape[0]) * self.meta.noise_level
            )
            query += noise

            new_support_set = []
            for s in support_set:
                noise = (
                    self.random_state.randn(s.shape[0]) * self.meta.noise_level
                )
                new_support_set.append(s + noise)
            support_set = np.array(new_support_set)

        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s

    @classmethod
    def name(cls) -> str:
        return "dm_cheat_counterfactual"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional, Tuple
from pathlib import Path
import pytorch_lightning as pl
from torch.utils.data import DataLoader
from itertools import cycle
from tqdm import tqdm

from meta.data.batch import TripletBatch
from meta.data.sampling import SuperSamplingTripletDataset
from meta.datasets.registry import register_data_module
from meta.datasets import get_data_module


@register_data_module
class SuperDataModule(pl.LightningDataModule):
    """
    A data module which provides a collection of GluonTS datasets.

     Parameters
    ----------

    dataset_names_train: str
        Names of the GluonTS datasets used for training.
    dataset_names_val: str
        Names of the GluonTS datasets used for validation.
    dataset_names_test: str
        Names of the GluonTS datasets used for testing.
    num_queries: int
        Number of queries used for the same support set.
    support_set_size: int
        Number of time series in the support set.
    prediction_length: Optional[int]
        Number of predicted time steps.
    context_length_multiple: int
        Context length is prediction length x ``context_length_multiple``.
    support_length_multiple: int
        Support time series length is prediction length x ``support_length_multiple``.
    dataset_sampling: str
         To generate training samples, first a dataset is chosen from the pool of training datasets.
         Then a sample is chosen from the chosen dataset. Training datasets can either be sampled uniformly
         or weighted w.r.t. to the total number of observations.
    data_dir: Path
        Path to data.
    standardize: bool
        If each time series is standardized.
    batch_size_train: int
        Batch size during training.
    batch_size_val_test: int
        Batch size during validation and testing.
    num_workers: int
        Number of workers.
    seed: int
        The random seed.
    catch22_train: bool
        If catch22 similarity measure is used for support set selection during training.
    catch22_val_test: bool
        If catch22 similarity measure is used for support set selection during validation and testing.
    cheat: float
        Chance of hiding the ground truth in the support set.
    """

    def __init__(
        self,
        dataset_names_train: str,
        dataset_names_val: str,
        dataset_names_test: str,
        num_queries: int,
        support_set_size: int,
        prediction_length: Optional[int] = None,
        context_length_multiple: int = 4,
        support_length_multiple: int = 4,
        dataset_sampling: str = "weighted",
        data_dir: Path = Path.home() / ".mxnet" / "gluon-ts",
        standardize: bool = True,
        batch_size_train: int = 128,
        batch_size_val_test: int = 1000,
        num_workers: int = 0,
        seed: Optional[int] = None,
        catch22_train: bool = False,
        catch22_val_test: bool = False,
        cheat: float = 0.0,
    ):
        super().__init__()
        self.dataset_names_train = dataset_names_train.split(",")
        self.dataset_names_val = dataset_names_val.split(",")
        self.dataset_names_test = dataset_names_test.split(",")
        self.standardize = standardize

        self.dm_config = {
            "support_set_size": support_set_size,
            "prediction_length": prediction_length,
            "context_length_multiple": context_length_multiple,
            "support_length_multiple": support_length_multiple,
            "standardize": standardize,
            "data_dir": data_dir,
            "batch_size_val_test": batch_size_val_test,
            "num_workers": num_workers,
            "seed": seed,
            "catch22_train": catch22_train,
            "catch22_val_test": catch22_val_test,
        }
        self.data_modules_train = [
            get_data_module(
                name="dm_" + d_name,
                dataset_name=d_name,
                num_queries=num_queries,
                cheat=cheat,
                **self.dm_config,
            )
            for d_name in self.dataset_names_train
        ]
        self.data_modules_val = [
            get_data_module(
                name="dm_" + d_name,
                dataset_name=d_name,
                num_queries=1,
                cheat=float(cheat > 0),
                **self.dm_config,
            )
            for d_name in self.dataset_names_val
        ]
        self.data_modules_test = [
            get_data_module(
                name="dm_" + d_name,
                dataset_name=d_name,
                num_queries=1,
                cheat=float(cheat > 0),
                **self.dm_config,
            )
            for d_name in self.dataset_names_test
        ]
        self.batch_size_train = batch_size_train
        self.batch_size_val_test = batch_size_val_test
        self.dataset_sampling = dataset_sampling
        self.catch22_train = catch22_train
        self.catch22_val_test = catch22_val_test
        self.num_workers = num_workers

    @property
    def prediction_length(self) -> int:
        all_dms = (
            self.data_modules_train
            + self.data_modules_val
            + self.data_modules_test
        )
        return max([dm.prediction_length for dm in all_dms])

    def setup(self, stage: Optional[str] = None) -> None:
        for dm in tqdm(
            self.data_modules_train
            + self.data_modules_val
            + self.data_modules_test,
            desc="setup data loader",
        ):
            dm.setup()

    def train_dataloader(self) -> DataLoader[TripletBatch]:
        super_sampling_dataset = SuperSamplingTripletDataset(
            datasets=[
                dm.sampling_triplet_dataset("train")
                for dm in self.data_modules_train
            ],
            dataset_sampling=self.dataset_sampling,
        )
        return DataLoader(
            super_sampling_dataset,
            collate_fn=TripletBatch.collate,  # type: ignore
            batch_size=self.batch_size_train,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def val_dataloader(self) -> DataLoader[TripletBatch]:
        return list(
            [dm_val.val_dataloader() for dm_val in self.data_modules_val]
        )

    def test_dataloader(self) -> DataLoader[TripletBatch]:
        return list(
            [dm_test.test_dataloader() for dm_test in self.data_modules_test]
        )

    def get_log_batches(self, n_logging_samples: int) -> Tuple[TripletBatch]:
        its_train = cycle(
            list(
                [
                    iter(dm.sampling_triplet_dataset("train"))
                    for dm in self.data_modules_train
                ]
            )
        )
        its_val = cycle(
            list(
                [
                    iter(dm.sequential_triplet_dataset("val"))
                    for dm in self.data_modules_val
                ]
            )
        )

        def get_log_batch(its):
            samples = []
            for i in range(n_logging_samples):
                samples.append(next(next(its)))
            samples = sorted(
                samples, key=lambda triplet: triplet.query_past[0].dataset_name
            )
            return TripletBatch.collate(samples)

        return get_log_batch(its_train), get_log_batch(its_val)

    @classmethod
    def name(cls) -> str:
        return "dm_super"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from pathlib import Path
from copy import deepcopy
import json

import numpy as np
import pandas as pd
import torch as pt

from .dataset import (
    FourierDataset,
    BenchmarkDataset,
    DomAdaptDataset,
)
from .estimator import (
    AttentionEstimator,
    AdversarialEstimator,
    DomAdaptEstimator,
    AdversarialDomAdaptEstimator,
)
from .engine import (
    AttentionTrainer,
    AttentionEvaluator,
    DomAdaptTrainer,
    DomAdaptEvaluator,
    AdversarialDomAdaptTrainer,
    AdversarialDomAdaptEvaluator,
)
from .cli import (
    NestedArgumentParser,
    NestedNamespace,
)


_freqs = {
    "electricity": "H",
    "traffic": "H",
    "wiki10k": "D",
    "kaggle": "D",
}

_window_size = {
    "H": 24 * 8,
    "D": 7 * 5,
}

_horizons = {
    "H": 24,
    "D": 7,
}


def update_configs(
    configs: NestedNamespace,
    defaults: dict,
) -> NestedNamespace:
    for key, value in defaults.items():
        if isinstance(value, dict):
            if isinstance(getattr(configs, key), NestedNamespace):
                namespace = getattr(configs, key)
                update_configs(namespace, value)
            else:
                raise ValueError("Incompatible structures.")
        else:
            if isinstance(getattr(configs, key), NestedNamespace):
                raise ValueError("Incompatible structures.")
            elif getattr(configs, key) is None:
                setattr(configs, key, value)


def parse_configs(cmd=None):
    parser = NestedArgumentParser()

    data_configs = parser.add_argument_group("data")
    data_configs.add_argument("src_name", type=str)
    data_configs.add_argument("tgt_name", type=str)
    data_configs.add_argument("--n_src_instances_train", type=int)
    data_configs.add_argument("--n_tgt_instances_train", type=int)
    data_configs.add_argument("--n_src_instances_eval", type=int)
    data_configs.add_argument("--n_tgt_instances_eval", type=int)
    data_configs.add_argument("--src_eval_size", type=int)
    data_configs.add_argument("--tgt_eval_size", type=int)

    model_configs = parser.add_argument_group("model")
    model_configs.add_argument("--d_hidden", type=int)
    model_configs.add_argument("--n_layer", type=int)
    model_configs.add_argument("--window_size", nargs="+", type=int)
    model_configs.add_argument("--n_head", type=int)
    model_configs.add_argument("--n_enc_layer", type=int)
    model_configs.add_argument("--n_dec_layer", type=int)
    model_configs.add_argument("--n_disc_layer", type=int)
    model_configs.add_argument("--symmetric", action="store_true")
    model_configs.add_argument("--share_values", action="store_true")
    model_configs.add_argument("--tie_casts", action="store_true")
    model_configs.add_argument("--tie_layers", action="store_true")
    model_configs.add_argument("--dropout", type=float)
    model_configs.add_argument("--temperature", type=float)
    model_configs.add_argument(
        "--not_scale_input", action="store_false", dest="normalize_input"
    )
    model_configs.add_argument(
        "--not_scale_output", action="store_false", dest="rescale_output"
    )
    model_configs.add_argument(
        "--last_layer_loss_only", action="store_false", dest="layerwise_loss"
    )

    train_configs = parser.add_argument_group("train")
    train_configs.add_argument("--lr", type=float)
    train_configs.add_argument("--n_epochs", type=int)
    train_configs.add_argument("--nb_epoch", type=int)
    train_configs.add_argument("--batch_size", type=int, nargs="+")
    train_configs.add_argument("--max_grad_norm", type=float)
    train_configs.add_argument("--n_loader_workers", type=int)
    train_configs.add_argument("--betas", type=float, nargs="+")
    train_configs.add_argument("--weight_decay", type=float)
    train_configs.add_argument("--cuda_device", type=int)

    global_configs = parser.add_argument_group("global")
    global_configs.add_argument("--seed", type=int)

    args = parser.parse_args(cmd)
    return args


def get_dataset(
    src_name: str,
    tgt_name: str,
    n_src_instances_train: int,
    n_tgt_instances_train: int,
    n_src_instances_eval: int,
    n_tgt_instances_eval: int,
    src_eval_size: int,
    tgt_eval_size: int,
):
    src_dataset = BenchmarkDataset.create_dataset(
        dataset=src_name,
        freq=_freqs[src_name],
        n_instances_train=n_src_instances_train,
        n_instances_eval=n_src_instances_eval,
        eval_size=src_eval_size,
        window_size=_window_size[_freqs[src_name]],
    )
    tgt_dataset = BenchmarkDataset.create_dataset(
        dataset=tgt_name,
        freq=_freqs[tgt_name],
        n_instances_train=n_tgt_instances_train,
        n_instances_eval=n_tgt_instances_eval,
        eval_size=tgt_eval_size,
        train_size=tgt_eval_size,
        window_size=_window_size[_freqs[tgt_name]],
    )
    return src_dataset, tgt_dataset


def main():
    args = parse_configs()
    src_name = args.data.src_name
    tgt_name = args.data.tgt_name
    exp_name = f"{src_name[0].upper()}2{tgt_name[0].upper()}"
    with Path(__file__).parents[1].joinpath(
        f"dumps/{exp_name}/defaults.json"
    ).open("r") as f:
        defaults = json.load(f)
    update_configs(args, defaults)
    src_dataset, tgt_dataset = get_dataset(**vars(args.data))

    n_disc_layer = args.model.n_disc_layer
    delattr(args.model, "n_disc_layer")
    seed = getattr(args, "global").seed

    pt.manual_seed(seed)
    S1 = AttentionEstimator.from_configs(
        src_dataset.d_data,
        src_dataset.d_feats,
        horizon=_horizons[_freqs[src_name]],
        **vars(args.model),
    )
    args.train.log_dir = f"dumps/{exp_name}/src"
    S1_trainer = AttentionTrainer.from_configs(
        src_dataset,
        S1,
        **vars(args.train),
    )
    pt.cuda.manual_seed_all(seed)
    S1_trainer.fit()
    S1_evaluator = AttentionEvaluator.from_trainer(S1_trainer, "best")
    S1_evaluator.evaluate()
    print(S1_evaluator.metrics.test.value)

    pt.manual_seed(seed)
    T0 = AttentionEstimator.from_configs(
        tgt_dataset.d_data,
        tgt_dataset.d_feats,
        horizon=_horizons[_freqs[tgt_name]],
        **vars(args.model),
    )
    args.train.log_dir = f"dumps/{exp_name}/tgt"
    T0_trainer = AttentionTrainer.from_configs(
        tgt_dataset,
        T0,
        **vars(args.train),
    )
    pt.cuda.manual_seed_all(seed)
    T0_trainer.fit()
    T0_evaluator = AttentionEvaluator.from_trainer(T0_trainer, "best")
    T0_evaluator.evaluate()
    print(T0_evaluator.metrics.test.value)

    DA_dataset = DomAdaptDataset.from_domains(src_dataset, tgt_dataset)
    T1 = S1.create_twin_estimator(
        tgt_dataset.d_data,
        tgt_dataset.d_feats,
        horizon=_horizons[_freqs[tgt_name]],
    )
    DA = DomAdaptEstimator(S1, T1)
    args.train.log_dir = f"dumps/{exp_name}/da"
    DA_trainer = DomAdaptTrainer.from_configs(
        DA_dataset,
        DA,
        **vars(args.train),
    )
    DA_trainer.fit()
    DA_evaluator = DomAdaptEvaluator.from_trainer(DA_trainer, "best")
    DA_evaluator.evaluate()
    print(DA_evaluator.metrics.test.value)

    S1_evaluator = AttentionEvaluator.from_trainer(S1_trainer, "best")
    S2 = AdversarialEstimator.from_base(S1, n_disc_layer)
    T2 = S2.create_twin_estimator(
        tgt_dataset.d_data,
        tgt_dataset.d_feats,
        horizon=_horizons[_freqs[tgt_name]],
    )
    ADV = AdversarialDomAdaptEstimator(S2, T2)
    args.train.log_dir = f"dumps/{exp_name}/adv"
    ADV_trainer = AdversarialDomAdaptTrainer.from_configs(
        DA_dataset,
        ADV,
        **vars(args.train),
    )
    ADV_trainer.fit()
    ADV_evaluator = AdversarialDomAdaptEvaluator.from_trainer(
        ADV_trainer, "best"
    )
    ADV_evaluator.evaluate()
    print(ADV_evaluator.metrics.test.value)


if __name__ == "__main__":
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import torch
import torch.nn.functional as F
from torch.distributions import (
    constraints,
    NegativeBinomial,
    Poisson,
    Distribution,
    TransformedDistribution,
    AffineTransform,
)
from torch.distributions.utils import broadcast_all, lazy_property

from .utils import broadcast_shape


class PiecewiseLinear(Distribution):
    def __init__(self, gamma, slopes, knot_spacings, validate_args=None):
        self.gamma = gamma
        self.slopes = slopes
        self.knot_spacings = knot_spacings

        self.b, self.knot_positions = PiecewiseLinear._to_orig_params(
            slopes=slopes, knot_spacings=knot_spacings
        )
        super(PiecewiseLinear, self).__init__(
            batch_shape=self.gamma.shape, validate_args=validate_args
        )

    @staticmethod
    def _to_orig_params(slopes, knot_spacings):
        # b: the difference between slopes of consecutive pieces
        b = slopes[..., 1:] - slopes[..., 0:-1]

        # Add slope of first piece to b: b_0 = m_0
        m_0 = slopes[..., 0:1]
        b = torch.cat((m_0, b), dim=-1)

        # The actual position of the knots is obtained by cumulative sum of
        # the knot spacings. The first knot position is always 0 for quantile
        # functions.
        knot_positions = torch.cumsum(knot_spacings, dim=-1) - knot_spacings

        return b, knot_positions

    @torch.no_grad()
    def sample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        u = torch.rand_like(self.gamma.expand(shape))

        sample = self.quantile(u)

        if len(sample_shape) == 0:
            sample = sample.squeeze(0)

        return sample

    def quantile(self, level):
        return self.quantile_internal(level, dim=0)

    def quantile_internal(self, x, dim=None):
        if dim is not None:
            gamma = self.gamma.unsqueeze(dim=dim if dim == 0 else -1)
            knot_positions = self.knot_positions.unsqueeze(dim)
            b = self.b.unsqueeze(dim)
        else:
            gamma, knot_positions, b = self.gamma, self.knot_positions, self.b

        x_minus_knots = x.unsqueeze(-1) - knot_positions

        quantile = gamma + (b * F.relu(x_minus_knots)).sum(-1)

        return quantile

    def log_prob(self, value):
        return -self.crps(value)

    def cdf(self, x):
        gamma, b, knot_positions = self.gamma, self.b, self.knot_positions

        quantiles_at_knots = self.quantile_internal(knot_positions, dim=-2)

        # Mask to nullify the terms corresponding to knots larger than l_0,
        # which is the largest knot (quantile level) such that the quantile
        # at l_0, s(l_0) < x.
        mask = torch.le(quantiles_at_knots, x.unsqueeze(-1))

        slope_l0 = (b * mask).sum(-1)

        # slope_l0 can be zero in which case a_tilde = 0.
        # The following is to circumvent an issue where the
        # backward() returns nans when slope_l0 is zero in the where
        slope_l0_nz = torch.where(
            slope_l0 == 0.0, torch.ones_like(x), slope_l0
        )

        a_tilde = torch.where(
            slope_l0 == 0.0,
            torch.zeros_like(x),
            (x - gamma + (b * knot_positions * mask).sum(-1)) / slope_l0_nz,
        )

        return torch.clamp(a_tilde, min=0.0, max=1.0)

    def crps(self, x):
        gamma, b, knot_positions = self.gamma, self.b, self.knot_positions

        a_tilde = self.cdf(x)

        max_a_tilde_knots = torch.max(a_tilde.unsqueeze(-1), knot_positions)

        knots_cubed = torch.pow(knot_positions, 3.0)
        coeff = (
            (1.0 - knots_cubed) / 3.0
            - knot_positions
            - torch.square(max_a_tilde_knots)
            + 2 * max_a_tilde_knots * knot_positions
        )

        return (
            (2 * a_tilde - 1) * x
            + (1 - 2 * a_tilde) * gamma
            + (b * coeff).sum(-1)
        )


class TransformedPiecewiseLinear(TransformedDistribution):
    def __init__(self, base_distribution, transforms):
        super().__init__(base_distribution, transforms)

    def crps(self, x):
        scale = 1.0

        for transform in reversed(self.transforms):
            assert isinstance(
                transform, AffineTransform
            ), "Not an AffineTransform"
            x = transform.inv(x)
            scale *= transform.scale

        p = self.base_dist.crps(x)
        return p * scale



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
from sklearn.metrics import auc


def bounded_pr_auc(
    precisions: np.ndarray, recalls: np.ndarray, lower_bound: float = 0
) -> float:
    """
    Bounded PR AUC --> AUC when recall > lower_bound.

    Parameters
    ----------
    precisions : np.ndarray
        precisions of different thresholds
    recalls : np.ndarray
        recalls of different thresholds
    lower_bound : float
        lower bound of recalls

    Returns
    -------
    bounded PR-AUC : float
    """
    sorted_recalls, sorted_precisions = zip(
        *sorted(zip(recalls, precisions), key=lambda x: (x[0], x[1]))
    )
    arg_num = np.argmax(np.array(sorted_recalls) >= lower_bound)
    pr_auc = auc(sorted_recalls[arg_num:], sorted_precisions[arg_num:])
    return pr_auc
